def _get_submodule(model, key):
    parent = model.get_submodule(".".join(key.split(".")[:-1]))
    target_name = key.split(".")[-1]
    target = model.get_submodule(key)
    return parent, target_name, target


def _set_submodule(parent, target_name, target):
    setattr(parent, target_name, target)


def transform_submodule(model, key, fn):
    parent, target_name, target = _get_submodule(model, key)
    transformed_target = fn(target)
    _set_submodule(parent, target_name, transformed_target)



import torch
import re
import warnings
from collections import OrderedDict
from inspect import signature, _empty
from types import SimpleNamespace
from contextlib import ExitStack, contextmanager
from copy import deepcopy
from typing import (
    Any,
    Dict,
    Optional,
    Union,
    Sequence,
    get_args,
    get_origin,
    get_type_hints,
)
from functools import partial
from dmx.compressor.modeling.nn import *
from dmx.compressor.fx.transform import (
    substitute_transform,
    prepare_tracing_inputs,
    make_compiler_graph,
)
from dmx.compressor.fx.transformer import get_op_set_from
from dmx.compressor.utils.fx.visualize_graph import visualize_graph


class DmxModelMixin:
    transformed: bool
    additional_dmx_aware_mappings: Dict
    _gm: Optional[torch.fx.GraphModule]  # current gm
    _gms: Dict  # stores {sig: gm} pairs
    _dmx_configuration_queue: List  # stores (config, rules) to be applied
    _monitoring_records: Optional[Dict]  # stored monitored submodule inputs/outputs
    _runtime_records: Optional[Dict]  # stored monitored submodule run times

    def _apply_config(self, config: Optional[Union[dict, str]], *rules):
        if config is not None:
            if isinstance(config, str):
                config = DmxConfig.from_yaml(config)

            for n, m in self.named_dmx_modules():
                if n in config:
                    m.configure(config[n])

        for _r in rules:
            _r.apply_to(self)

    def configure(self, config: Optional[Union[dict, str]], *rules):
        r"""
        Configure Dmx-specific numerics/sparsity/logics

        NOTE: only staticly declared DmxModule(s) are to be transformed

        Args:
            config (Optional[Union[DmxConfig, str]]): DmxConfig to be used for transformation.
            *rules (List[DmxConfigRule]): variable length of list of configuration rules on top of config.

        Returns:
            Returns the transformed model

        """
        self._dmx_configuration_queue.append((config, rules))
        if self.transformed:
            self._apply_config(config, *rules)
        return self

    transform = configure  # NOTE: to be deprecated

    @property
    def op_set(self):
        r"Returns a set of unique ops present in the model"
        return set.union(*[get_op_set_from(_m) for _m in self._gms.values()])

    @property
    def dmx_config(self):
        r""" "Returns the DmxConfig object for the model"""
        return DmxConfig.from_model(self, freeze=True)

    @property
    def dmx_module_names(self):
        r""" "Returns a list of module names listed in a dmx_config"""
        return self.dmx_config.module_names

    def named_dmx_modules(self):
        r""" "Returns a list of named modules that are dmx configurable"""

        #To guard against some cases where this function is called on
        #submodules
        if hasattr(self,'_gms') and self._gms is not None:
            all_modules = []
            for gm in self._gms.values():
                new_modules = [(f'_gm.{n}',m) for n,m in gm.named_modules() if \
                               is_configurable(m) and \
                               f'_gm.{n}' not in [x[0] for x in all_modules]]
                all_modules.extend(new_modules)
            return (x for x in all_modules)
        else:
            return ((n, m) for n, m in self.named_modules() if is_configurable(m))


    def freeze(self, config_file="./config.yaml"):
        """
        A function that stores the state and ops format of the model to a config file

        Args:
            config_file (Optional[str]): Path to store the config file. Defaults to "./config.yaml".
        """
        DmxConfig.from_model(self, freeze=True).to_yaml(config_file)

    def thaw(self, config_file="./config.yaml"):
        """
        A function that transforms the model in place from a config file.

        Args:
            config_file (Optional[str]): Path of config file to transform the model from. Defaults to "./config.yaml".
        """
        self.configure(config_file)

    def print_model_tree(self, include_type=False):
        """
        A function that prints out the tree structure of a model

        Args:
            include_type (bool): include the type of modules in the print out if True
        """
        from dmx.compressor.utils.visualization import print_model_tree

        print_model_tree(self, include_type)

    def fold_weights_and_biases(self):
        """
        A function that applies the ops the weights and biases using the corresponding formats.
        """
        for _, m in self.named_dmx_modules():
            m.fold_weight_and_bias()

    def check_dim_consistency(self) -> bool:
        """
        A function that checks format dimension consistency and sparseness dimension consistency for all applicable dmx modules in the model

        Returns:
            True if all dimensions are consistent for all applicable modules.
        """
        return all(
            m.check_format_dim_consistency() and m.check_sparseness_dim_consistency()
            for _, m in self.named_dmx_modules()
        )

    def to_baseline_mode(self):
        from dmx.compressor import config_rules

        # we can clear the _dmx_configuration_queue as baseline config rule will overwrite everything
        self._dmx_configuration_queue = []
        self.configure(None, *config_rules.BASELINE)

    def to_basic_mode(self, sbfp_weight_storage=False):
        """
        Configures a transformed DmxModel to the BASIC mode on dmx hardware.

        Returns:
            The configured model.
        """
        from dmx.compressor import config_rules

        self.to_baseline_mode()
        self.configure(None, *config_rules.BASIC)
        if sbfp_weight_storage:
            self.configure(None, *config_rules.SBFP_WEIGHT_STORAGE)
            self.forward_weight_hypernets()

    @contextmanager
    def keep_dmx_config(self):
        _old_dmx_config_queue = deepcopy(self._dmx_configuration_queue)
        yield self
        for _config, _rules in _old_dmx_config_queue:
            self._apply_config(_config, *_rules)
        self._dmx_configuration_queue = _old_dmx_config_queue

    @contextmanager
    def counting_flops(self, zero: bool = True):
        with ExitStack() as stack:
            yield [
                stack.enter_context(m.counting_flops(zero))
                for _, m in self.named_dmx_modules()
            ]

    @staticmethod
    def _save_specific_layers_state_dict_and_register_urls(
        specific_layers: Sequence[DmxModule],
        save_checkpoint_to: Optional[str],
    ):
        if save_checkpoint_to is not None:
            for _m in specific_layers:
                _m.save_state_dict_and_register_url(parent_dir=save_checkpoint_to)

    @contextmanager
    def monitoring(
        self,
        submodules_to_monitor: List[str] = [],
        save_checkpoint_to: Optional[str] = None,
    ):
        self._monitoring_records = {_sm: [] for _sm in submodules_to_monitor}
        with ExitStack() as stack:
            yield [
                stack.enter_context(
                    self._gm.get_submodule(_sm).monitoring(
                        self._monitoring_records[_sm]
                    ),
                )
                for _sm in submodules_to_monitor
            ]

    def get_monitoring_records(self, submodules_to_monitor: List[str] = []):
        _rec = self._monitoring_records
        self._monitoring_records = None

        return _rec

    @contextmanager
    def measure_runtimes(self, device, submodules_to_measure: List[str] = []):
        self._runtime_records = {_sm: [] for _sm in submodules_to_measure}
        with ExitStack() as stack:
            yield [
                stack.enter_context(
                    self._gm.get_submodule(_sm).measuring_runtime(
                        self._runtime_records[_sm], device
                    ),
                )
                for _sm in submodules_to_measure
            ]

    def get_runtime_records(self):
        _rec = self._runtime_records
        self._runtime_records = None

        return _rec

    @staticmethod
    def _add_transformed_gm(_model, args, kwargs):
        if hasattr(_model, "old_forward"):
            _model.forward = _model.old_forward
        input_names, concrete_args, dummy_inputs = prepare_tracing_inputs(
            _model, args, kwargs
        )
        _model._gm = substitute_transform(
            _model,
            input_names=input_names,
            concrete_args=concrete_args,
            dummy_inputs=dummy_inputs,
            additional_mappings=_model.additional_dmx_aware_mappings,
        )
        DmxModel.post_process_gm(_model, args, kwargs)

    @staticmethod
    def _get_transformed_forward(_model):
        if hasattr(_model, "old_forward"):
            _model.forward = _model.old_forward
        _mod_signature = signature(_model.forward)
        _output_cls = _mod_signature.return_annotation
        if get_origin(_output_cls) is Union:  # NOTE: this is error-prone
            transformer_output_cls = None
            for output_type in get_args(_output_cls):
                # filter out typing classes, eg: typing.Tuple, as they will throw error with issubclass
                if not hasattr(output_type, "__origin__") and issubclass(
                    output_type, transformers.modeling_utils.ModelOutput
                ):
                    transformer_output_cls = output_type
                    break
            _output_cls = transformer_output_cls
        elif (
            _output_cls is _empty
            or hasattr(_output_cls, "__origin__")
            or not issubclass(_output_cls, transformers.modeling_utils.ModelOutput)
        ):
            _output_cls = None

        _model._output_cls = _output_cls
        _forward = (
            (
                lambda *_args, **_kwargs: DmxModel._wrap_output_cls(
                    _model._gm.forward(*_args, **_kwargs), _output_cls
                )
            )
            if _model._output_cls is not None
            else _model._gm.forward
        )
        return _forward

    @staticmethod
    def _wrap_output_cls(outputs, cls):
        attribute_types = get_type_hints(cls)
        for key, expected_type in attribute_types.items():
            if key in outputs:
                if (
                    expected_type
                    and isinstance(expected_type, type)
                    and issubclass(
                        expected_type, transformers.utils.generic.ModelOutput
                    )
                ):
                    outputs[key] = DmxModel._wrap_output_cls(
                        outputs[key], expected_type
                    )  # Recursively wrap sub-objects

        return cls(**outputs)

    @staticmethod
    def is_same_signature(_model, args, kwargs):

        tracing_args, tracing_kwargs = _model.tracing_kwargs
        # if kwargs has different keys, need to retrace
        if tracing_kwargs.keys() != kwargs.keys():
            return False
        # if args has different length, need to retrace
        if len(tracing_args) != len(args):
            return False
        # comparing args values between tracing args and new args
        for a, new_a in zip(tracing_args, args):
            # if bool argument has different values, need to retrace
            if isinstance(new_a, bool) and a != new_a:
                return False
            # if one in None and other is not none, need to retrace
            if not isinstance(new_a, bool) and (new_a is None) != (a is None):
                return False
        # comparing kwargs values between tracing kwarg and new kwargs
        for k in kwargs.keys():
            # if bool argument has different values, need to retrace
            if isinstance(kwargs[k], bool) and kwargs[k] != tracing_kwargs[k]:
                return False
            # if type of argument does not match, need to retrace
            if type(kwargs[k]) != type(tracing_kwargs[k]):
                return False
            # if kwargs is DynamicCache and shape is different (i.e: one is empty), need to retrace.
            # NOTE: to be deprecated
            if isinstance(kwargs[k], transformers.DynamicCache) and len(
                kwargs[k]
            ) != len(tracing_kwargs[k]):
                return False
        return True

    def forward_weight_hypernets(self):
        for _, _m in self.named_dmx_modules():
            if hasattr(_m, "weight"):
                _ = _m._weight

    @staticmethod
    def deepcopy_args(args):
        def deepcopy_element(element):
            if isinstance(element, torch.Tensor):
                return element.detach().clone()
            elif isinstance(element, tuple):
                return type(element)(deepcopy_element(e) for e in element)
            elif isinstance(element, list):
                return type(element)([deepcopy_element(e) for e in element])
            elif isinstance(element, dict):
                return type(element)(
                    {k: deepcopy_element(v) for k, v in element.items()}
                )
            else:
                return deepcopy(element)

        return deepcopy_element(args)

    @staticmethod
    def post_process_gm(_model, args, kwargs):
        # some inputs were removed from input names due to None or bool, we want to add it back to maintain original input signature
        placeholders_needed = list(
            signature(_model.forward).bind(*args, **kwargs).arguments.keys()
        )
        node_list = _model._gm.graph.nodes
        i = 0
        for node in node_list:
            if i >= len(placeholders_needed):
                break
            while (
                i < len(placeholders_needed) and node.target != placeholders_needed[i]
            ):
                with _model._gm.graph.inserting_before(node):
                    _model._gm.graph.placeholder(placeholders_needed[i])
                    i += 1
            i += 1
        _model._gm.recompile()

    @classmethod
    def create_submod_transform_forward(
        cls,
        model: torch.nn.Module,
        submod_name: str,
        additional_dmx_aware_mappings=None,
    ):
        submod = model.get_submodule(submod_name)
        if DmxModelMixin not in submod.__class__.__bases__:
            submod.__class__.__bases__ += (DmxModelMixin,)
        submod._gm = None
        submod.transformed = False
        submod.config = model.config
        submod.additional_dmx_aware_mappings = additional_dmx_aware_mappings

        def temp_forward(_m, *_args, **_kwargs):
            _is_training = _m.training
            if not _m.transformed or not DmxModel.is_same_signature(_m, _args, _kwargs):
                if not model.transformed:
                    raise Exception(
                        "model forward needs to be called before submodule forward."
                    )
                warnings.warn("Submodule transformation triggered")
                _m.tracing_kwargs = (
                    DmxModel.deepcopy_args(_args),
                    DmxModel.deepcopy_args(_kwargs),
                )
                # because some args and kwargs are changed in place in forward, we need to keep a copy fo the _args and _kwargs
                forward_args, forward_kwargs = DmxModel.deepcopy_args(
                    _args
                ), DmxModel.deepcopy_args(_kwargs)

                DmxModel._add_transformed_gm(_m, _args, _kwargs)
                _m._forward = DmxModel._get_transformed_forward(_m)
                _m.transformed = True

                for n, _ in _m.named_dmx_modules():
                    dmx_mod_name = n[4:]  # first 4 chars are _gm.
                    dmx_mod_name = "_gm." + submod_name + "." + dmx_mod_name

                    # handling nested attr for setattr
                    pre, _, post = n.rpartition(".")
                    if pre:
                        parent = submod.get_submodule(pre)
                    else:
                        parent = submod
                    setattr(parent, post, model.get_submodule(dmx_mod_name))

                _m.train(_is_training)
                _m.transformed_forward = partial(temp_forward, _m)
                return _m._forward(*forward_args, **forward_kwargs)

            return _m._forward(*_args, **_kwargs)

        submod.old_forward = submod.forward
        submod.transformed_forward = partial(temp_forward, submod)

    @staticmethod
    def to_signature_key(_m, _args, _kwargs):
        sig = signature(_m.old_forward)
        inputs = signature(_m.old_forward).bind(*_args, **_kwargs).arguments
        sig_key = tuple()
        for k, v in sig.parameters.items():
            if k not in inputs:
                value = v.default
            else:
                value = inputs[k]
            if isinstance(value, bool) or value is None:
                sig_key += (value,)
            elif isinstance(value, transformers.DynamicCache) and len(value) == 0:
                sig_key += (None,)
            else:
                sig_key += (repr(type(value)),)
        return sig_key


class DmxModel(DmxModelMixin):
    @classmethod
    def from_torch(
        cls, model: torch.nn.Module, additional_dmx_aware_mappings=None
    ) -> torch.nn.Module:
        if not isinstance(model, cls):
            _cls = model.__class__
            model.class_for_deserialization = _cls
            model.__class__ = _cls.__class__("Dmx" + _cls.__name__, (_cls, cls), {})
        model.additional_dmx_aware_mappings = additional_dmx_aware_mappings
        model.transformed = False
        model._gm = None
        model._gms = {}
        model._monitoring_records = None
        model._runtime_records = None
        from dmx.compressor import config_rules

        model._dmx_configuration_queue = [(None, config_rules.BASELINE)]

        def temp_forward(_m, *_args, **_kwargs):
            _is_training = _m.training
            sig_key = DmxModel.to_signature_key(_m, _args, _kwargs)
            if not _m.transformed or not DmxModel.is_same_signature(_m, _args, _kwargs):
                warnings.warn("Model transformation triggered")
                _m.tracing_kwargs = (
                    DmxModel.deepcopy_args(_args),
                    DmxModel.deepcopy_args(_kwargs),
                )
                if _m.transformed:
                    current_config = _m.dmx_config
                else:
                    current_config = {}
                if sig_key in _m._gms:
                    warnings.warn("Reusing graph module from past")
                    _m._gm = _m._gms[sig_key]
                else:
                    # because some args and kwargs are changed in place in forward, we need to keep a copy fo the _args and _kwargs
                    transform_args, transform_kwargs = DmxModel.deepcopy_args(
                        _args
                    ), DmxModel.deepcopy_args(_kwargs)
                    DmxModel._add_transformed_gm(_m, transform_args, transform_kwargs)
                    _m._gms[sig_key] = _m._gm
                
                _m._forward = DmxModel._get_transformed_forward(_m)

                _m.transformed = True
                # configure the model
                for _config, _rules in _m._dmx_configuration_queue:
                    _m._apply_config(_config, *_rules)
                _m._apply_config(current_config)
                _m.train(_is_training)
                _m.forward = partial(temp_forward, _m)

                return _m._forward(*_args, **_kwargs)

            return _m._forward(*_args, **_kwargs)

        model.old_forward = model.forward
        model.forward = partial(temp_forward, model)

        return model

    def visualize_graph(self, out_file="graph"):
        if not self.transformed:
            raise RuntimeError(
                "A forward pass is needed before model can be visualized!"
            )
        tracing_args, tracing_kwargs = self.tracing_kwargs
        inputs = tuple(
            signature(self.old_forward)
            .bind(*tracing_args, **tracing_kwargs)
            .arguments.values()
        )
        visualize_graph(self._gm, inputs, out_file)

    def make_compiler_graphs(self):
        if not hasattr(self, "compiler_graphs"):
            self._compiler_graphs = dict()
        for key, gm in self._gms.items():
            if key not in self._compiler_graphs:
                self._compiler_graphs[key] = make_compiler_graph(gm)


class DmxConfig(dict):
    r"""
    This is a dict of Dmx-specific configurations for a DmxModel
    This defines the 'states' to be optimized
    """

    @classmethod
    def from_model(cls, model: torch.nn.Module, freeze=False):
        """
        A function that stores state and ops format of the model in a DmxConfig object

        Args:
            model (Model): Target model for creating the DmxConfig

        Returns:
            A DmxConfig object that stores state and ops format of the model in a DmxConfig object
        """
        return cls({n: m.dmx_config(freeze) for n, m in model.named_dmx_modules()})

    @classmethod
    def from_yaml(cls, fname):
        """
        A function that creates a DmxConfig object from a yaml file

        Args:
            fname (str): file path of the yaml file

        Returns:
            A DmxConfig object vreated from yaml file
        """
        from dmx.compressor.utils.io import load_config_file

        return cls(load_config_file(fname))

    def to_yaml(self, fname):
        """
        A function that stores the content of a DmxConfig object to a yaml file

        Args:
            fname (str): file path of the yaml file
        """
        from dmx.compressor.utils.io import save_config_file

        save_config_file(dict(self), fname)

    @property
    def module_names(self):
        """
        Returns the module names in the DmxConfig object
        """
        return self.keys()


class DmxConfigRule(SimpleNamespace):
    r"""
    This is a rule that specifies how to transform from DmxConfig to DmxConfig
    This defines the 'action' in the state space

    Args:
        module_types (Optional[Tuple]): Types of modules to apply transformation. Defaults to empty Tuple.
        name_re (Optional[str]): String for matching module name patterns. Defaults to empty str.
        module_config (Optional[DmxModuleConfig]): DmxModuleConfig that specifies the ops formats for a module. Defaults to empty DmxModuleConfig.

    Attributes:
        module_types (Tuple): Types of modules to apply transformation.
        name_rule (Pattern): Pattern for matching module names.
        module_config (DmxModuleConfig): DmxModuleConfig that specifies the ops formats for a module.
    """

    def __init__(
        self,
        module_types=(),
        name_re: str = "",
        module_config: DmxModuleConfig = DmxModuleConfig(),
    ) -> None:
        assert all([issubclass(mt, DmxModule) for mt in module_types])
        self.module_types = module_types
        self.name_rule = re.compile(name_re)
        self.module_config = module_config

    def names_in(self, model_or_config: Union[torch.nn.Module, DmxConfig]):
        """
        Creates a list of module names where the modules are in self.module_types and the names match with self.name_rule.

        Args:
            model_or_config (Union[torch.nn.Module, DmxConfig]): torch.nn.Module or DmxConfig to create the name of modules for.

        Returns:
            A list of module names
        """
        config = (
            model_or_config
            if isinstance(model_or_config, DmxConfig)
            else DmxConfig.from_model(model_or_config, freeze=True)
        )
        return [
            n
            for n in config.module_names
            if any(
                issubclass(config[n]["instance_of"], mod_type)
                for mod_type in self.module_types
            )
            and self.name_rule.match(n)
        ]

    def apply_to(self, model_or_config: Union[torch.nn.Module, DmxConfig]):
        """
        A function that sets format of ops according to self.module_config for modules selected by self.module_types and
        self.name_rule on a model or DmxConfig

        Args:
            model_or_config (Union[Model, DmxConfig]): Model or DmxConfig to apply transformation on.
        """
        target_module_names = self.names_in(model_or_config)
        if isinstance(model_or_config, torch.nn.Module):
            for n, m in model_or_config.named_dmx_modules():
                if n in target_module_names and isinstance(m, self.module_types):
                    m.configure(self.module_config)
        else:
            config = model_or_config
            for n in target_module_names and any(
                issubclass(config[n]["instance_of"], mod_type)
                for mod_type in self.module_types
            ):
                config[n].update(self.module_config)


# alias for backward compatibility, to be deprecated
DmxTransformation = DmxConfigRule


class DmxPipelineMixin:
    def configure(
        self,
        dmx_config_dict: Dict[str, Optional[DmxConfig]],
        dmx_transformation_dict: Optional[Dict[str, Optional[DmxConfigRule]]] = None,
    ) -> None:
        for _n, _m in self.named_dmx_models():
            tr = [] if dmx_transformation_dict is None else dmx_transformation_dict[_n]
            _m.transform(dmx_config_dict[_n], *tr)

    transform = configure  # NOTE: to be deprecated

    @contextmanager
    def counting_flops(self, zero: bool = True) -> None:
        with ExitStack() as stack:
            yield [
                stack.enter_context(_m.counting_flops(zero))
                for _, _m in self.named_dmx_models()
            ]

    def eval(self):
        for _m in self.model_dict.values():
            _m.eval()

    @property
    def dmx_config_dict(self) -> Dict[str, DmxConfig]:
        return {n: m.dmx_config for n, m in self.named_dmx_models()}

    def named_dmx_modules(self):
        return (
            (f"{_model_name}.{_module_name}", _module)
            for _model_name, _model in self.named_dmx_models()
            for _module_name, _module in _model.named_dmx_modules()
        )

    def get_model_by_name(self, model_name: str) -> torch.nn.Module:
        return self.model_dict[model_name]


class DmxSimplePipeline(DmxPipelineMixin):
    def __init__(
        self,
        model_dict: OrderedDict,
        preproc=torch.nn.Identity(),
        postproc=torch.nn.Identity(),
        hf: bool = False,
        concrete_args_dict: Optional[Dict] = None,
    ) -> None:
        self.model_dict = model_dict
        self.preproc = preproc
        self.postproc = postproc
        self.hf = hf
        self.concrete_args_dict = concrete_args_dict

    def __call__(self, *args, **kwargs):
        _out = self.postproc(
            torch.nn.Sequential(*[m for m in self.model_dict.values()])(
                self.preproc(*args, **kwargs)
            )
        )
        return _out

    def named_dmx_models(self):
        r"Returns a generator of named DmxModel instances"
        return ((n, m) for n, m in self.model_dict.items() if isinstance(m, Model))

    def to(self, torch_device: Optional[Union[str, torch.device]] = None):
        if torch_device is not None:
            for _n, _m in self.named_dmx_models():
                _m.to(torch_device)
        return self

    @property
    def op_set(self):
        return set.union(*[get_op_set_from(_m) for _m in self.model_dict.values()])


class Model(DmxSimplePipeline):
    r"""
    This is a backward-compatible placeholder for legacy models.
    It is not recommended to use this container; instead, use DmxSimplePipeline directly.
    TODO: to be deprecated

    """

    def __init__(
        self,
        body,
        head=torch.nn.Identity(),
        tail=torch.nn.Identity(),
        hf: bool = False,
        concrete_args: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            model_dict=dict(model=DmxModel.from_torch(body, concrete_args)),
            preproc=head,
            postproc=tail,
            hf=hf,
            concrete_args_dict=dict(model=concrete_args),
        )
        self.body = self.model_dict["model"]
        self.head = self.preproc
        self.tail = self.postproc

    @property
    def op_set(self):
        r"Returns a set of unique ops present in the model"
        return set.union(*[get_op_set_from(_m) for _m in self._gms.values()])



from typing import Dict, Optional, Set
import torch.fx as fx
from torch.fx.proxy import Proxy
import re
import torch

from dmx.compressor.modeling import nn as dmxnn
import transformers

dmx_aware_mapping = {
    "torch.nn.modules.sparse.Embedding": dmxnn.Embedding,
    "torch.nn.modules.linear.Linear": dmxnn.Linear,
    "torch.nn.modules.conv.Conv1d": dmxnn.Conv1d,
    "torch.nn.modules.conv.Conv2d": dmxnn.Conv2d,
    "torch.nn.modules.pooling.AdaptiveAvgPool2d": dmxnn.AdaptiveAvgPool2d,
    "torch.nn.modules.pooling.MaxPool2d": dmxnn.MaxPool2d,
    "torch.nn.modules.batchnorm.BatchNorm2d": dmxnn.BatchNorm2d,
    "torch.nn.modules.normalization.LayerNorm": dmxnn.LayerNorm,
    "torch.nn.modules.dropout.Dropout": dmxnn.Dropout,
    "torch.nn.modules.activation.Softmax": dmxnn.Softmax,
    "torch.nn.modules.activation.ReLU": dmxnn.ReLU,
    "torch.nn.modules.activation.ReLU6": dmxnn.ReLU6,
    "torch.nn.modules.activation.SiLU": dmxnn.SiLU,
    "torch.nn.modules.activation.Tanh": dmxnn.Tanh,
    "torch.nn.modules.activation.GELU": dmxnn.GELU,
}

transformer_module_mapping = {
    "transformers.pytorch_utils.Conv1D": dmxnn.Linear,
    "transformers.activations.NewGELUActivation": dmxnn.NewGELU,
    "transformers.activations.GELUActivation": dmxnn.GELU,
    "transformers.activations.FastGELUActivation": dmxnn.FastGELU,
    "transformers.activations.QuickGELUActivation": dmxnn.QuickGELU,
    "transformers.activations.ClippedGELUActivation": dmxnn.ClippedGELU,
    "transformers.models.bloom.modeling_bloom.BloomGelu": dmxnn.BloomGELU,
    "transformers.models.t5.modeling_t5.T5LayerNorm": dmxnn.RMSNorm,
    "transformers.models.llama.modeling_llama.LlamaRMSNorm": dmxnn.RMSNorm,
    "transformers.models.gemma.modeling_gemma.GemmaRMSNorm": dmxnn.GemmaRMSNorm,
    "transformers.models.mistral.modeling_mistral.MistralRMSNorm": dmxnn.RMSNorm,
    "transformers.models.llama.modeling_llama.LlamaRotaryEmbedding": dmxnn.LlamaRotaryEmbedding,
}

dmx_aware_functional_mappings = {
    "torch.nn.functional.relu": dmxnn.ReLU,
    "torch.nn.functional.relu6": dmxnn.ReLU6,
    "torch.nn.functional.silu": dmxnn.SiLU,
    "torch.nn.functional.tanh": dmxnn.Tanh,
    "torch.nn.functional.gelu": dmxnn.GELU,
    "torch.nn.functional.softmax": dmxnn.Softmax,
    "torch.nn.functional.dropout": dmxnn.Dropout,
    "torch.matmul": dmxnn.ActActMatMul,
    "torch.exp": dmxnn.Exp,
    "torch.bmm": dmxnn.ActActMatMul,
    "torch.baddbmm": dmxnn.BAddBMM,
    "torch.nn.functional.scaled_dot_product_attention": dmxnn.ScaledDotProductAttention,
}

transformer_function_mapping = {
    "transformers.models.llama.modeling_llama.apply_rotary_pos_emb": dmxnn.ApplyRotaryPosEmb,
    "transformers.models.gemma.modeling_gemma.apply_rotary_pos_emb": dmxnn.ApplyRotaryPosEmb,
    "transformers.models.mistral.modeling_mistral.apply_rotary_pos_emb": dmxnn.ApplyRotaryPosEmb,
}

dmx_aware_method_mapping = {
    "matmul": dmxnn.ActActMatMul,
    "baddbmm": dmxnn.BAddBMM,
    "exp": dmxnn.Exp,
    "bmm": dmxnn.ActActMatMul,
    "add": dmxnn.ResAdd,
    "mul": dmxnn.Mul,
}

dmx_aware_mapping.update(transformer_module_mapping)
dmx_aware_functional_mappings.update(transformer_function_mapping)
for f_key in list(dmx_aware_functional_mappings.keys()):
    new_key = repr(eval(f_key))
    dmx_aware_functional_mappings[new_key] = dmx_aware_functional_mappings.pop(f_key)
dmx_aware_functional_mappings["<built-in function add>"] = dmxnn.ResAdd
dmx_aware_functional_mappings["<built-in function matmul>"] = dmxnn.ActActMatMul
dmx_aware_functional_mappings["<built-in function mul>"] = dmxnn.Mul


def process_args(args):
    """
    A function that goes through each arg in args and removes the proxy wrapper.

    Args:
        args (Tuple): args for which proxy wrappers should be removed

    Returns:
        A tuple of args where each element has the proxy wrapper removed

    """
    new_args = ()
    for arg in args:
        if isinstance(arg, Proxy):
            new_args += (arg.node,)
        else:
            new_args += (arg,)
    return new_args


def process_kwargs(kwargs):
    """
    A function that goes through each kwarg in kwargs and removes the proxy wrapper.

    Args:
        kwargs (Dict): kwargs for which proxy wrappers should be removed

    Returns:
        A dictionary of kwargs where each element has the proxy wrapper removed

    """
    new_kwargs = dict()
    for k in kwargs.keys():
        if isinstance(kwargs[k], Proxy):
            new_kwargs[k] = kwargs[k].node
        else:
            new_kwargs[k] = kwargs[k]
    return new_kwargs


def get_name_for_func_nodes(
    candidate: str, used_names: Set[str], base_count: Dict[str, int]
):
    """Get the unique name for functional nodes

    Arguments:
        candidate (str): used as the basis for the unique name, relevant to the user.
        used_names (Set[str]): A Set of names already used for nodes
        base_count (Dict[str, int]): A dict counting number of names sharing the same candidate base
    """
    _illegal_char_regex = re.compile("[^0-9a-zA-Z_]+")
    _name_regex = re.compile(r"^([a-zA-Z_][0-9a-zA-Z_]*?)(?:_(\d+))?$")

    match = _name_regex.match(candidate)
    if match is None:
        # delete all characters that are illegal in a Python identifier
        candidate = _illegal_char_regex.sub("_", candidate)

        if not candidate:
            candidate = "_unnamed"

        if candidate[0].isdigit():
            candidate = f"_{candidate}"

        match = _name_regex.match(candidate)
        assert match is not None

    base, num = match.group(1, 2)
    if num is None or candidate in used_names:
        num = base_count.get(candidate, 0)
    else:
        num = int(num)

    while candidate in used_names:
        num += 1
        candidate = f"{base}_{num}"

    return candidate


def _extract_ops(gm: torch.fx.GraphModule):
    r"""
    A generator that gathers ops from a GraphModule
    """
    for _op in gm.graph.nodes:
        if _op.op == "call_module":
            yield _op.name, (gm.get_submodule(_op.target).__class__)
        if _op.op == "call_method":
            yield _op.name, getattr(torch.Tensor, _op.target)
        if _op.op == "call_function":
            yield _op.name, _op.target


def get_op_set_from(gm: Optional[torch.fx.GraphModule]):
    r"""
    Returns a set of ops from a GraphModule
    """
    return set(t for _, t in _extract_ops(gm)) if gm is not None else set()


def gap_analysis(source, target):
    r"Print commonality and differences between new and existing op sets"
    known = target.intersection(source)
    unknown = target.difference(source)
    print(
        f"""Known ops (n = {len(known)}):
    {known}
    """
    )
    print(
        f"""Unknown ops (n = {len(unknown)}):
    {unknown}
    """
    )



#!/usr/bin/env python3
#
import logging
import math
import torch
from torch import fx
from .. import numerical
from .. import sparse
from .. import modeling
from .transformer.utils import transformer_function_mapping, transformer_module_mapping

from torch.fx.node import Argument, Node, Target
from typing import Any, Callable, Dict, Optional, Tuple, Type, List, Union
import transformers.utils.fx as fx_hf
import transformers

from transformers.utils.fx import (
    HFTracer,
    get_concrete_args,
)
from torch.fx.graph_module import GraphModule
from transformers.modeling_utils import PreTrainedModel
from contextlib import contextmanager
import inspect


class DmxHFTracer(HFTracer):
    """
    Custom HFTracer where definition of leaf nodes
    """

    def __init__(
        self,
        autowrap_modules=(math,),
        autowrap_functions=tuple(eval(k) for k in transformer_function_mapping.keys()),
    ):
        super().__init__(
            autowrap_modules=autowrap_modules, autowrap_functions=autowrap_functions
        )

    def is_leaf_module(self, m: torch.nn.Module, module_qualified_name: str) -> bool:
        """
        A method to specify whether a given ``nn.Module`` is a "leaf" module.

        Leaf modules are the atomic units that appear in
        the IR, referenced by ``call_module`` calls. By default,
        Modules in the PyTorch standard library namespace (torch.nn)
        are leaf modules. All other modules are traced through and
        their constituent ops are recorded, unless specified otherwise
        via this parameter.

        Args:
            m (Module): The module being queried about
            module_qualified_name (str): The path to root of this module. For example,
                if you have a module hierarchy where submodule ``foo`` contains
                submodule ``bar``, which contains submodule ``baz``, that module will
                appear with the qualified name ``foo.bar.baz`` here.
        Returns:
            True if m is a leaf module
        """
        is_leaf = isinstance(
            m,
            (
                numerical.CastTo,
                sparse.Sparsify,
                transformers.pytorch_utils.Conv1D,
            ),
        )
        is_leaf = is_leaf or (
            isinstance(m, tuple(eval(k) for k in transformer_module_mapping.keys()))
        )
        is_leaf = is_leaf or (
            isinstance(m, modeling.nn.DmxModule) and not m.is_compound
        )

        is_leaf = is_leaf or (
            (
                m.__module__.startswith("torch.nn")
                or m.__module__.startswith("torch.ao.nn")
                or m.__module__.startswith("transformers.activations")
            )
            and not isinstance(m, torch.nn.Sequential)
        )
        return is_leaf


@contextmanager
def disable_hooked_forward(model):
    """
    This context manager disables hooks inserted by accelerate for fx tracing, and adds it back to the old forward upon exiting the context manager
    """
    for m in model.modules():
        if hasattr(m, "_old_forward"):
            m.hooked_forward = m.forward
            m.forward = m._old_forward
    if hasattr(model, "_old_forward"):
        model.hooked_forward = model.forward
        model.forward = model._old_forward
    yield
    for m in model.modules():
        if hasattr(m, "hooked_forward"):
            m.forward = m.hooked_forward
    if hasattr(model, "hooked_forward"):
        model.old_forward = model.hooked_forward


def hf_symbolic_trace(
    model: PreTrainedModel,
    input_names: Optional[List[str]] = None,
    concrete_args: Optional[Dict[str, Any]] = None,
    tracer_cls: Type[DmxHFTracer] = DmxHFTracer,
    dummy_inputs: Optional[Dict[str, Any]] = None,
) -> GraphModule:
    """
    Performs symbolic tracing on a huggingface model.

    Args:
        model ([`PretrainedModel`]):
            The model to trace.
        input_names (`List[str]`, *optional*):
            The names of the inputs of the traced model. If unset, model.dummy_inputs.keys() are used instead.
        disable_check (`bool`, *optional*, defaults to `False`):
            If `True`, no check is done before trying to trace the model, this is mostly usesul for debugging purposes.
        tracer_cls (`Type[HFTracer]`, *optional*, defaults to `HFTracer`):
            The tracer class to use for instantiating the tracer. If unset, `HFTracer` is used instead.

    Returns:
        `torch.fx.GraphModule`: A GraphModule constructed by recording operations seen while tracing the model.
        `torch.fx.Tracer`: The tracer used for tracing the model.

    Example:

        ```python
        from dmx.compressor.fx.tracer import hf_symbolic_trace

        traced_model,tracer = hf_symbolic_trace(model, input_names=["input_ids", "attention_mask", "token_type_ids"])
        ```
    """

    if not hasattr(model, "config"):
        model.config = None
    with disable_hooked_forward(model):
        if input_names is None:
            input_names = model.dummy_inputs.keys()

        input_names = list(input_names)
        if concrete_args:
            new_args = get_concrete_args(model, input_names)
            for key, value in new_args.items():
                if key not in concrete_args:
                    concrete_args[key] = value
            concrete_args.update()
        else:
            concrete_args = get_concrete_args(model, input_names)
        sig = inspect.signature(
            model.forward if isinstance(model, torch.nn.Module) else model
        )
        # set dummy_input for params without default
        for param in sig.parameters.values():
            if param.name in dummy_inputs:
                continue
            if param.default is inspect.Parameter.empty:
                dummy_inputs[param.name] = None

        # Tracing.
        tracer = tracer_cls()
        traced_graph = tracer.trace(
            model,
            concrete_args=concrete_args,
            dummy_inputs=dummy_inputs,
        )
        traced = GraphModule(model, traced_graph)
    return (traced, tracer)


def symbolic_trace(
    root: Union[torch.nn.Module, Callable[..., Any]],
    concrete_args: Optional[Dict[str, Any]] = None,
) -> GraphModule:
    """
    Symbolic tracing API

    Given an ``nn.Module`` or function instance ``root``, this function will return a ``GraphModule``
    constructed by recording operations seen while tracing through ``root`` and the tracer used to trace the model.

    ``concrete_args`` allows you to partially specialize your function, whether it's to remove control flow or data structures.

    For example::

        def f(a, b):
            if b == True:
                return a
            else:
                return a*2

    FX can typically not trace through this due to the presence of control
    flow. However, we can use `concrete_args` to specialize on the value of
    `b` to trace through this::

        f = fx.symbolic_trace(f, concrete_args={'b': False})
        assert f(3, False)  == 6

    Note that although you can still pass in different values of `b`, they will be ignored.

    We can also use `concrete_args` to eliminate data-structure handling from
    our function. This will use pytrees to flatten your input. To avoid
    overspecializing, pass in `fx.PH` for values that shouldn't be
    specialized. For example::

        def f(x):
            out = 0
            for v in x.values():
                out += v
            return out
        f = fx.symbolic_trace(f, concrete_args={'x': {'a': fx.PH, 'b': fx.PH, 'c': fx.PH}})
        assert f({'a': 1, 'b': 2, 'c': 4}) == 7


    Args:
        root (Union[torch.nn.Module, Callable]): Module or function to be traced and converted
            into a Graph representation.
        concrete_args (Optional[Dict[str, any]]): Inputs to be partially specialized

    Returns:
        GraphModule: a Module created from the recorded operations from ``root``.
        Tracer: the tracer used for tracing the model
    """
    tracer = fx.Tracer()
    graph = tracer.trace(root, concrete_args)
    name = (
        root.__class__.__name__ if isinstance(root, torch.nn.Module) else root.__name__
    )
    return (GraphModule(tracer.root, graph, name), tracer)


# Referenced from https://pytorch.org/docs/stable/_modules/torch/ao/quantization/quantize_fx.html#convert_fx
class Scope(object):
    """Scope object that records the module path and the module type
    of a module. Scope is used to track the information of the module
    that contains a Node in a Graph of GraphModule.

    Example:
        class Sub(torch.nn.Module):
            def forward(self, x):
                # This will be a call_method Node in GraphModule,
                # scope for this would be (module_path="sub", module_type=Sub)
                return x.transpose(1, 2)

        class M(torch.nn.Module):
            def __init__(self):
                self.sub = Sub()

            def forward(self, x):
                # This will be a call_method Node as well,
                # scope for this would be (module_path="", None)
                x = x.transpose(1, 2)
                x = self.sub(x)
                return x

    Args:
        module_path (str): String describing the path to the module
        module_type (Any): type of the module

    Attributes:
        module_path (str): String describing the path to the module
        module_type (Any): type of the module


    """

    def __init__(self, module_path: str, module_type: Any):
        super().__init__()
        self.module_path = module_path
        self.module_type = module_type


class ScopeContextManager(object):
    """A context manager to track the Scope of Node during symbolic tracing.
    When entering a forward function of a Module, we'll update the scope information of
    the current module, and when we exit, we'll restore the previous scope information.

    Args:
        scope (Scope): Scope object to store the module details
        current_module (torch.nn.Module): Current module object
        current_module_path (str): String path to current module

    Attributes:
        prev_module_type (Any): Type of the previous module
        prev_module_path (str): String path to previous module
        scope (Scope): Scope object to store the module details
    """

    def __init__(
        self, scope: Scope, current_module: torch.nn.Module, current_module_path: str
    ):
        super().__init__()
        self.prev_module_type = scope.module_type
        self.prev_module_path = scope.module_path
        self.scope = scope
        self.scope.module_path = self.prev_module_path + "__" + current_module_path
        self.scope.module_type = type(current_module)

    def __enter__(self):
        return

    def __exit__(self, *args):
        """Restore information of the previous module on exit."""
        self.scope.module_path = self.prev_module_path
        self.scope.module_type = self.prev_module_type
        return


class QuantTracer(fx.Tracer):
    """
    Customed tracer with scope manager and returns a flat GraphModule

    Attributes:
        scope (Scope): Scope object to record the path and type of a module
        node_name_to_scope (Dict): Dictionary that maps node name to scope
        record_stack_traces (bool): Not in use yet
    """

    def __init__(self) -> None:
        super().__init__()
        self.scope = Scope("model", None)
        self.node_name_to_scope: Dict[str, Tuple[str, type]] = {}
        self.record_stack_traces = True

    def is_leaf_module(self, m: torch.nn.Module, module_qualified_name: str) -> bool:
        """
        A method to specify whether a given ``nn.Module`` is a "leaf" module.

        Leaf modules are the atomic units that appear in
        the IR, referenced by ``call_module`` calls. By default,
        Modules in the PyTorch standard library namespace (torch.nn)
        are leaf modules. All other modules are traced through and
        their constituent ops are recorded, unless specified otherwise
        via this parameter.

        Args:
            m (Module): The module being queried about
            module_qualified_name (str): The path to root of this module. For example,
                if you have a module hierarchy where submodule ``foo`` contains
                submodule ``bar``, which contains submodule ``baz``, that module will
                appear with the qualified name ``foo.bar.baz`` here.

        Returns:
            True if m is a leaf module
        """

        is_leaf = isinstance(
            m,
            (
                numerical.CastTo,
                sparse.Sparsify,
            ),
        )
        return is_leaf

    def call_module(
        self,
        m: torch.nn.Module,
        forward: Callable[..., Any],
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
    ) -> Any:
        """
        Method that specifies the behavior of this ``Tracer`` when it encounters
        a call to an ``nn.Module`` instance.

        By default, the behavior is to check if the called module is a leaf module
        via ``is_leaf_module``. If it is, emit a ``call_module`` node referring to
        ``m`` in the ``Graph``. Otherwise, call the ``Module`` normally, tracing through
        the operations in its ``forward`` function.

        Args:
            m (Module): The module for which a call is being emitted
            forward (Callable): The forward() method of the ``Module`` to be invoked
            args (Tuple): args of the module callsite
            kwargs (Dict): kwargs of the module callsite

        Returns:
            The return value from the Module call. In the case that a ``call_module``
            node was emitted, this is a ``Proxy`` value. Otherwise, it is whatever
            value was returned from the ``Module`` invocation.
        """
        module_qualified_name = self.path_of_module(m)
        # Creating scope with information of current module
        # scope will be restored automatically upon exit
        with ScopeContextManager(self.scope, m, module_qualified_name):
            logger = logging.getLogger(__name__)
            logger.info("path:", self.scope.module_path)
            logger.info("type:", self.scope.module_type)
            return super().call_module(m, forward, args, kwargs)

    def create_node(
        self,
        kind: str,
        target: Target,
        args: Tuple[Argument, ...],
        kwargs: Dict[str, Argument],
        name: Optional[str] = None,
        type_expr: Optional[Any] = None,
    ) -> Node:
        """
        Inserts a graph node given target, args, kwargs, and name.

        Args:
            op (str): the opcode for this Node. One of 'call_function', 'call_method', 'get_attr',
                'call_module', 'placeholder', or 'output'. The semantics of these opcodes are
                described in the ``Graph`` docstring.
            args (Optional[Tuple[Argument, ...]]): is a tuple of arguments to this node.
            kwargs (Optional[Dict[str, Argument]]): the kwargs of this Node
            name (Optional[str]): an optional string name for the ``Node``.
                This will influence the name of the value assigned to in the
                Python generated code.
            type_expr (Optional[Any]): an optional type annotation representing the
                Python type the output of this node will have.

        Returns:
            The newly-created and inserted node.
        """
        node = super().create_node(kind, target, args, kwargs, name, type_expr)
        self.node_name_to_scope[node.name] = (
            self.scope.module_path,
            self.scope.module_type,
        )
        return node


class HFQuantTracer(fx_hf.HFTracer):
    """
    Customed tracer with scope manager for HuggingFace

    Attributes:
        scope (Scope): Scope object to record the path and type of a module
        node_name_to_scope (Dict): Dictionary that maps node name to scope
        record_stack_traces (bool): Not in use yet
    """

    def __init__(self) -> None:
        super().__init__()
        self.scope = Scope("model", None)
        self.node_name_to_scope: Dict[str, Tuple[str, type]] = {}
        self.record_stack_traces = True

    def is_leaf_module(self, m: torch.nn.Module, module_qualified_name: str) -> bool:
        """
        A method to specify whether a given ``nn.Module`` is a "leaf" module.

        Leaf modules are the atomic units that appear in
        the IR, referenced by ``call_module`` calls. By default,
        Modules in the PyTorch standard library namespace (torch.nn)
        are leaf modules. All other modules are traced through and
        their constituent ops are recorded, unless specified otherwise
        via this parameter.

        Args:
            m (Module): The module being queried about
            module_qualified_name (str): The path to root of this module. For example,
                if you have a module hierarchy where submodule ``foo`` contains
                submodule ``bar``, which contains submodule ``baz``, that module will
                appear with the qualified name ``foo.bar.baz`` here.
        Returns:
            True if m is a leaf module
        """
        is_leaf = isinstance(
            m,
            (
                numerical.CastTo,
                sparse.Sparsify,
            ),
        )
        return is_leaf

    def call_module(
        self,
        m: torch.nn.Module,
        forward: Callable[..., Any],
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
    ) -> Any:
        """
        Method that specifies the behavior of this ``Tracer`` when it encounters
        a call to an ``nn.Module`` instance.

        By default, the behavior is to check if the called module is a leaf module
        via ``is_leaf_module``. If it is, emit a ``call_module`` node referring to
        ``m`` in the ``Graph``. Otherwise, call the ``Module`` normally, tracing through
        the operations in its ``forward`` function.

        Args:
            m (Module): The module for which a call is being emitted
            forward (Callable): The forward() method of the ``Module`` to be invoked
            args (Tuple): args of the module callsite
            kwargs (Dict): kwargs of the module callsite

        Returns:
            The return value from the Module call. In the case that a ``call_module``
            node was emitted, this is a ``Proxy`` value. Otherwise, it is whatever
            value was returned from the ``Module`` invocation.
        """
        module_qualified_name = self.path_of_module(m)
        # Creating scope with information of current module
        # scope will be restored automatically upon exit
        with ScopeContextManager(self.scope, m, module_qualified_name):
            logger = logging.getLogger(__name__)
            logger.info("path:", self.scope.module_path)
            logger.info("type:", self.scope.module_type)
            return super().call_module(m, forward, args, kwargs)

    def create_node(
        self,
        kind: str,
        target: Target,
        args: Tuple[Argument, ...],
        kwargs: Dict[str, Argument],
        name: Optional[str] = None,
        type_expr: Optional[Any] = None,
    ) -> Node:
        """
        Inserts a graph node given target, args, kwargs, and name.

        Args:
            op (str): the opcode for this Node. One of 'call_function', 'call_method', 'get_attr',
                'call_module', 'placeholder', or 'output'. The semantics of these opcodes are
                described in the ``Graph`` docstring.
            args (Optional[Tuple[Argument, ...]]): is a tuple of arguments to this node.
            kwargs (Optional[Dict[str, Argument]]): the kwargs of this Node
            name (Optional[str]): an optional string name for the ``Node``.
                This will influence the name of the value assigned to in the
                Python generated code.
            type_expr (Optional[Any]): an optional type annotation representing the
                Python type the output of this node will have.

        Returns:
            The newly-created and inserted node.
        """
        node = super().create_node(kind, target, args, kwargs, name, type_expr)
        self.node_name_to_scope[node.name] = (
            self.scope.module_path,
            self.scope.module_type,
        )
        return node



## This test file serves the purpose of testing the inheritance of castTo states after DmxModel retransformation
import torch
from dmx.compressor.modeling import DmxModel, DmxConfigRule
from dmx.compressor.modeling import nn as dmxnn
from dmx.compressor.advanced_recipe import (
    DmxQuantizerCalibrationHyperparams,
    DmxModuleQuantizerCalibrationHyperparams,
    DmxQuantizerCalibrationRecipe,
)
from copy import deepcopy

torch.manual_seed(0)


class Net(torch.nn.Module):
    def __init__(self, in_dim, out_dim):
        super().__init__()
        self.ln1 = torch.nn.Linear(in_dim, 128)
        self.act = torch.nn.ReLU()
        self.ln2 = torch.nn.Linear(128, out_dim)

    def forward(self, x, y=None):
        x = self.ln1(x)
        x = self.act(x)
        x = self.ln2(x)
        if y is not None:
            return x.T @ y
        return x


in_dim = 10
out_dim = 5
batch = 2

x = torch.rand((batch, in_dim))
y = torch.rand((batch, in_dim))


def test_retransformation_after_calib():
    """
    scales and zero points should be same while observer and fakequant states should be different
    """
    net = Net(in_dim, out_dim)
    model = DmxModel.from_torch(net)
    with torch.no_grad():
        model(x)
    format = "XP[8,0](CSN)"
    rules = (
        DmxConfigRule(
            module_types=(dmxnn.Linear,),
            module_config=dict(
                input_formats=[format],
                weight_format=format,
            ),
        ),
    )
    model.configure(None, *rules)
    target_layers = {
        n: m for n, m in model.named_dmx_modules() if isinstance(m, (dmxnn.Linear,))
    }
    hp_gen = lambda _model: {
        _m: DmxModuleQuantizerCalibrationHyperparams(
            inputs={"input_cast": DmxQuantizerCalibrationHyperparams()},
            weight=DmxQuantizerCalibrationHyperparams(),
        )
        for _, _m in _model.named_dmx_modules()
        if isinstance(_m, dmxnn.Linear)
    } 
    with DmxQuantizerCalibrationRecipe(hp_gen).applied_to(model), torch.no_grad():
        model(x)
        target_layers_copy = {n: deepcopy(m) for n, m in target_layers.items()}
    model(y, y)
    # check all scales and zerpoints are the same
    if any(
        (
            m.input_casts.input_cast.scale[0]
            != target_layers_copy[n].input_casts.input_cast.scale[0]
            or m.input_casts.input_cast.zero_point[0]
            != target_layers_copy[n].input_casts.input_cast.zero_point[0]
        )
        for n, m in target_layers.items()
    ):
        assert False, "scale and zero point changed after retransformation!"

    # check all fake_quant_enabled and observer_enabled are different
    if any(
        (
            m.input_casts.input_cast.fake_quant_enabled
            == target_layers_copy[n].input_casts.input_cast.fake_quant_enabled
            or m.input_casts.input_cast.observer_enabled
            == target_layers_copy[n].input_casts.input_cast.observer_enabled
        )
        for n, m in target_layers.items()
    ):
        assert (
            False
        ), "fake_quant_enabled and observer_enabled incorrect after retransformation!"


def test_retransformation_during_calib():
    """
    scales and zero points should be different while observer and fakequant states should be same
    """
    net = Net(in_dim, out_dim)
    model = DmxModel.from_torch(net)
    with torch.no_grad():
        model(x)
    format = "XP[8,0](CSN)"
    rules = (
        DmxConfigRule(
            module_types=(dmxnn.Linear,),
            module_config=dict(
                input_formats=[format],
                weight_format=format,
            ),
        ),
    )
    model.configure(None, *rules)
    target_layers = {
        n: m for n, m in model.named_dmx_modules() if isinstance(m, (dmxnn.Linear,))
    }
    hp_gen = lambda _model: {
        _m: DmxModuleQuantizerCalibrationHyperparams(
            inputs={"input_cast": DmxQuantizerCalibrationHyperparams()},
            weight=DmxQuantizerCalibrationHyperparams(),
        )
        for _, _m in _model.named_dmx_modules()
        if isinstance(_m, dmxnn.Linear)
    } 
    with DmxQuantizerCalibrationRecipe(hp_gen).applied_to(model), torch.no_grad():
        model(x)
        target_layers_copy = {n: deepcopy(m) for n, m in target_layers.items()}
        model(y, y)

        # check not all scales and zerpoints are the same
        if all(
            (
                m.input_casts.input_cast.scale[0]
                == target_layers_copy[n].input_casts.input_cast.scale[0]
                and m.input_casts.input_cast.zero_point[0]
                == target_layers_copy[n].input_casts.input_cast.zero_point[0]
            )
            for n, m in target_layers.items()
        ):
            assert (
                False
            ), "scale and zero point did not changed when calibrating after retransformation!"

        # check all fake_quant_enabled and observer_enabled are the same
        if any(
            (
                m.input_casts.input_cast.fake_quant_enabled
                != target_layers_copy[n].input_casts.input_cast.fake_quant_enabled
                or m.input_casts.input_cast.observer_enabled
                != target_layers_copy[n].input_casts.input_cast.observer_enabled
            )
            for n, m in target_layers.items()
        ):
            assert (
                False
            ), "fake_quant_enabled and observer_enabled incorrect after retransformation!"



import torch
from dmx.compressor.modeling import nn as dmxnn


def test_exp():
    net = dmxnn.Exp()
    x = torch.rand(8, 32)
    output_before = torch.exp(x)
    output = net(x)
    assert (output_before - output).abs().sum() <= 1e-5, True



#!/usr/bin/env python3
import torch
import torch.fx as fx
from typing import Any
from graphviz import Digraph
from dmx.compressor.fx import NodeDictTransformer, QuantTracer
from dmx.compressor.utils.fx.interpreter import MetadataInterpreter


def visualize_graph(
    model: torch.nn.Module, input: torch.Tensor, file_name="graph", tracer=QuantTracer()
) -> Any:
    """Saves the graph to file_name if given and returns the pygraph object
    Example use cases:
    visualizing non fx transformed:
        visualize_graph(torch.nn.Linear(64,64),torch.rand(1,64))
    visualizing fx transformed:
        net = DmxModel.from_torch(torch.nn.Linear(64,64))
        visualize_graph(net,torch.rand(1,64))
    Note: make sure inputs follow the same order of kwargs as the module signature, remember to install pygraphviz by sudo apt install graphviz
    """
    if not isinstance(model, fx.GraphModule):
        graph = tracer.trace(model)
        model = fx.GraphModule(tracer.root, graph)
    nodeDict = NodeDictTransformer(model).transform()
    gi = MetadataInterpreter(model, nodeDict)
    with torch.no_grad():
        if isinstance(input, tuple):
            gi.run(*input)
        else:
            gi.run(input)
    node_attr = dict(
        style="filled",
        shape="box",
        align="center",
        fontsize="12",
        height="0.1",
        fontname="monospace",
    )
    pygraph = Digraph(
        node_attr=node_attr,
        graph_attr=dict(layout="dot"),
        edge_attr=dict(penwidth="2", fontsize="15"),
    )
    pygraph.node(str(id("start")), label="start", fillcolor="white")
    pygraph.node(str(id("end")), label="end", fillcolor="white")
    for info_nodes in gi.nodes:
        pygraph.node(*(info_nodes.args), **(info_nodes.kwargs))
    for info_eges in gi.edges.values():
        if len(info_eges.args) == 3:
            pygraph.edge(*(info_eges.args), **(info_eges.kwargs))

    try:
        pygraph.render(filename=file_name)
    except RuntimeError:
        print(
            "Pygraphviz not installed from root! Please install by 'sudo apt install graphviz'"
        )



from typing import Any
import warnings
import torch
import torch.nn as nn
from parse import parse
from torch.autograd import Function
from torch.nn.modules.lazy import LazyModuleMixin
from torch.nn.parameter import UninitializedParameter
import transformers
from dmx.compressor.utils.visualization import mask2braille

__ALL__ = [
    "Sparseness",
    "WeightSparseMixin",
    "Dense",
    "TopK",
    "BlockTopK",
    "Sparsify",
    "SparsificationManager",
]


class Sparseness(Function):
    r"""
    This class inherits from torch.autograd.Function, since when we use masking
    operations, we may have different forward and backward passes.  Child
    classes to implement `get_mask()` and `from_shorthand()` method.
    """

    blocked: bool

    def __init__(self, mask_gradient=False):
        super().__init__()
        self.mask_gradient = torch.as_tensor(mask_gradient)

    def __str__(self) -> str:
        raise NotImplementedError

    def get_mask(self, *input: Any):
        raise NotImplementedError

    @staticmethod
    def backward(ctx, grad_output):
        # Propagrate None gradients to the parameters.
        # We'll need to run ctx.save_for_backward to save the two variables
        # mask_gradient and mask during the forward pass.
        mask_gradient, mask = ctx.saved_tensors[:2]
        if mask_gradient.item():
            return grad_output * mask, None
        else:
            return grad_output, None

    @classmethod
    def from_shorthand(cls, sh: str):
        if sh.startswith("DENSE"):
            return Dense.from_shorthand(sh)
        elif sh.startswith("TOPK"):
            return TopK.from_shorthand(sh)
        elif sh.startswith("BTOPK"):
            return BlockTopK.from_shorthand(sh)
        elif sh.startswith("BERN"):
            return Bernoulli.from_shorthand(sh)
        else:
            raise ValueError(f"unrecognized sparseness shorthand: {sh}")


class Dense(Sparseness):
    r"""
    This is a dummy sparsity whose `forward` and `backward` are equivalent to an
    identity function.
    """

    blocked = False

    def __init__(self, mask_gradient=False):
        super().__init__(mask_gradient)
        self.mask_gradient = mask_gradient
        self.density = 1.0

    def get_mask(self, score):
        return None

    @classmethod
    def from_shorthand(cls, sh: str):
        return cls()

    def __str__(self) -> str:
        return "Dummy sparseness: no pruning"

    def __repr__(self) -> str:
        return "DENSE"

    def __reduce__(self):
        return (self.__class__, (self.mask_gradient,))


class TopK(Sparseness):
    r"""
    Fine-grain unstructured sparsity with top-K scored entries non-zero
    """

    blocked = False

    def __init__(self, density=0.5, mask_gradient=False):
        super().__init__(mask_gradient)
        assert 0 <= density <= 1.0, "density has to be between 0 and 1"
        self.density = density

    @staticmethod
    def forward(ctx, score, params):
        (
            mask_gradient,
            density,
        ) = params
        _score = score.view(-1)
        idx = torch.argsort(_score, dim=0)[: int(score.numel() * (1.0 - density))]
        mask = (
            torch.ones_like(_score, device=score.device)
            .scatter_(dim=0, index=idx, value=0)
            .view_as(score)
        )
        ctx.save_for_backward(mask_gradient, mask)
        return mask

    def get_mask(self, score):
        params = (
            self.mask_gradient,
            self.density,
        )
        return TopK.apply(score, params)

    @classmethod
    def from_shorthand(cls, sh: str):
        conf = parse("TOPK{{{density:f}}}({mask_grad:l})", sh)
        return cls(
            density=conf["density"],
            mask_gradient=conf["mask_grad"] == "M",
        )

    def __str__(self) -> str:
        return f"Global TopK sparseness: density = {self.density}"

    def __repr__(self) -> str:
        return f"TOPK{{{self.density}}}({'M' if self.mask_gradient else 'U'})"


class BlockTopK(Sparseness):
    r"""
    Fine-grain structured sparsity with K non-zeros out of `block_size` elements
    long `block_dim`.
    """

    blocked = True

    def __init__(self, K=4, block_size=8, block_dim=-1, mask_gradient=False):
        super().__init__(mask_gradient)
        assert 0 < K <= block_size, "N and M must be positive and N no greater than M"
        self.K = K
        self.block_size = block_size
        self.block_dim = block_dim
        self.density = self.K / self.block_size

    @staticmethod
    def forward(ctx, score, params):
        mask_gradient, K, block_size, block_dim = params
        assert (
            score.shape[block_dim] % block_size == 0
        ), f"score has size {score.shape[block_dim]} at dimension {block_dim}, not a multiple of block size {block_size}"
        _score = score.transpose(block_dim, -1)
        score_shape = _score.shape
        _score = _score.reshape(-1, block_size)
        idx = torch.argsort(_score, dim=1)[:, : int(block_size - K)]
        mask = (
            torch.ones_like(_score, device=score.device)
            .scatter_(dim=1, index=idx, value=0)
            .reshape(score_shape)
            .transpose_(block_dim, -1)
        )
        ctx.save_for_backward(mask_gradient, mask)
        return mask

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output, None

    def get_mask(self, score):
        params = (self.mask_gradient, self.K, self.block_size, self.block_dim)
        return BlockTopK.apply(score, params)

    @classmethod
    def from_shorthand(cls, sh: str):
        conf = parse("BTOPK{{{K:d}:{block_size:d},{block_dim:d}}}({mask_grad:l})", sh)
        return cls(
            K=conf["K"],
            block_size=conf["block_size"],
            block_dim=conf["block_dim"],
            mask_gradient=conf["mask_grad"] == "M",
        )

    def __str__(self) -> str:
        return f"Block TopK sparseness: pattern = {self.K}:{self.block_size}, block dimension = {self.block_dim}"

    def __repr__(self) -> str:
        return f"BTOPK{{{self.K}:{self.block_size},{self.block_dim}}}({'M' if self.mask_gradient else 'U'})"


class Bernoulli(Sparseness):
    r"""Bernoulli sampler for supermasking."""

    blocked = False

    def __init__(self, mask_gradient=False):
        super().__init__(mask_gradient)
        self.density = None

    @staticmethod
    def forward(ctx, score, params):
        # The scores need to be within [0, 1], otherwise there are bugs
        assert score.max() <= 1
        assert score.min() >= 0
        mask_gradient = params[0]
        mask = torch.bernoulli(score)
        ctx.save_for_backward(mask_gradient, mask)
        return mask

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output, None

    def get_mask(self, score):
        params = (self.mask_gradient,)
        return Bernoulli.apply(score, params)

    @classmethod
    def from_shorthand(cls, sh: str):
        return cls()

    def __str__(self) -> str:
        return "Bernoulli sparseness"

    def __repr__(self) -> str:
        return "BERN"


class Sparsify(nn.Module):
    r"""
    Sparsification module
    """

    def __init__(
        self,
        tensor_shape,
        sparseness="DENSE",
        backward_mode="STE",
        score_func=None,
    ):
        super().__init__()

        self.score = nn.Parameter(torch.rand(tensor_shape), requires_grad=True)
        self.configure(sparseness, backward_mode, score_func)
        self.update_mask(self.score)
        self.plastic = False

    def configure(self, sparseness=None, backward_mode=None, score_func=None):
        # sparseness pattern
        if sparseness is not None:
            if not isinstance(sparseness, Sparseness):
                sparseness = Sparseness.from_shorthand(sparseness)
            if not hasattr(self, "sparseness") or repr(sparseness) != repr(
                self.sparseness
            ):
                self.sparseness = sparseness
        # backward mode
        if backward_mode is not None:
            self.backward_mode = backward_mode
            self.enable_weight_gradient = backward_mode.lower() in {"ste", "joint"}
            self.enable_mask_gradient = backward_mode.lower() in {"supermask", "joint"}
        # score function: importance <- score_func(score, input)
        if score_func is not None:
            self.score_func = score_func
            # mark the module as plastic for rewiring in the next forward() call
            self.plastic = True

    def update_mask(self, score):
        self.mask = self.sparseness.get_mask(score)

    def forward(self, x):
        if not isinstance(self.sparseness, Dense):
            if self.plastic:
                score = self.score_func(self.score, x)
                self.plastic = False
            else:
                score = self.score
            self.update_mask(score)
            if self.training:
                x = x if self.enable_weight_gradient else x.detach()
                self.mask = (
                    self.mask if self.enable_mask_gradient else self.mask.detach()
                )
            x = x * self.mask
        return x

    @property
    def density(self) -> float:
        if self.sparseness.density is not None:
            return self.sparseness.density
        else:
            self.update_mask(self.score)
            return self.mask.data.sum() / self.mask.numel()

    def mask_str(self, dims, max_elems):
        return mask2braille(self.mask, dims, max_elems)

    def extra_repr(self):
        return f"sparseness = {self.sparseness.__repr__()}, backward_mode = {self.backward_mode}"


class LazySparsify(LazyModuleMixin, Sparsify):
    cls_to_become = Sparsify
    score: UninitializedParameter

    def __init__(
        self,
        sparseness="DENSE",
        backward_mode="STE",
        score_func=None,
    ) -> None:
        super().__init__(torch.Size([0]), sparseness, backward_mode, score_func)
        self.score = UninitializedParameter()
        self.configure(sparseness, backward_mode, score_func)

    def reset_parameters(self) -> None:
        if not self.has_uninitialized_params():
            nn.init.uniform_(self.score)

    def initialize_parameters(self, x: torch.Tensor) -> None:
        self.tensor_shape = x.shape
        if self.has_uninitialized_params():
            with torch.no_grad():
                if isinstance(self.sparseness, Dense):
                    self.score.materialize(torch.Size([0]))
                else:
                    self.score.materialize(self.tensor_shape)
                    self.reset_parameters()


class SparsificationManager:
    r"""
    This is a sparsification management class
    Similar to Scheduler that manages parameter optimization through scheduler.step(),
    one can call sparsification_manager.step() to update underlying score.
    """

    def __init__(
        self,
        sparsify_modules,
        **kwargs,
    ):
        self.sparsify_modules = sparsify_modules

    def step(self, **kwargs):
        for sm in self.sparsify_modules:
            sm.configure(**kwargs)


class WeightSparseMixin:
    """Mixin for weight-sparse modules."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.init_sparsifier()

    def init_sparsifier(self):
        if isinstance(
            self,
            (
                nn.Linear,
                nn.Bilinear,
                nn.Embedding,
                nn.EmbeddingBag,
                nn.modules.conv._ConvNd,
                transformers.pytorch_utils.Conv1D,
            ),
        ):
            self.weight_sparsifier = LazySparsify()
        else:
            self.weight_sparsifier = None

    @property
    def effective_weight(self):
        return (
            self.weight
            if self.weight_sparsifier is None
            else self.weight_sparsifier(self.weight)
        )

    @property
    def weight_sparseness(self):
        return (
            (self.weight_sparsifier.sparseness)
            if self.weight_sparsifier is not None
            else None
        )

    @staticmethod
    def _check_sparseness_dim_consistency(sparseness: Sparseness, ch_axis: int) -> bool:
        return not sparseness.blocked or sparseness.block_dim == ch_axis

    def check_sparseness_dim_consistency(self) -> bool:
        _good = (
            self.weight_sparsifier is None
            or self._check_sparseness_dim_consistency(
                self.weight_sparsifier.sparseness, self.win_ch_axis
            )
        )
        if not _good:
            warnings.warn(
                f"layer's weight channel axis {self.win_ch_axis} might be inconsistent with sparseness {self.weight_sparseness}",
                RuntimeWarning,
            )
        return _good



from dataclasses import dataclass
from torch import Tensor
from typing import Union,List,TYPE_CHECKING,Dict,Any
from abc import ABC,abstractmethod
from contextlib import contextmanager


import dmx.compressor.modeling


tens_or_list_type = Union[Tensor,List[Tensor]]

@dataclass
class PluginLayerData:
    input_before_cast : tens_or_list_type
    input_after_cast : tens_or_list_type
    output_before_cast : tens_or_list_type
    output_after_cast : tens_or_list_type
    mod : "DmxModule"
    args : List[Any]
    kwargs : Dict[str,Any]

    
class PluginBase(ABC):
    @abstractmethod
    def process_model(self,model : "DmxModel"):
        '''
        Define here the model transformations that do not depend on a calibration input.
        For example, layernorm scale computation using SLANC

        '''
        pass

    @abstractmethod
    def process_layer(self, layer_data : PluginLayerData):
        '''
        Define here the transformations that depend on a calibration input.
        This function will be called for each layer (DmxModule) with the layer inputs and
        outputs. These come from the calibration data you call the model with. You have to call
        the model with this calibration data

        '''
        
        pass
    

class ActivatePlugins:
    def __init__(self,plugins : Union[PluginBase,List[PluginBase]]):
        if isinstance(plugins,PluginBase):
            plugins = [plugins]
        self.plugins = plugins

    @contextmanager
    def applied_to(self,model : "DmxModel"):
        for p in self.plugins:
            p.process_model(model)
        dmx.compressor.modeling.DmxModule.plugins = self.plugins
        yield
        dmx.compressor.modeling.DmxModule.plugins = []
        
    
                 



import torch
import pytest

from dmx.compressor.modeling.model import DmxModel, DmxConfigRule
from dmx.compressor.modeling import nn as dmxnn

# Add parent directory to Python path
import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../src')))


class Submod(torch.nn.Module):
    def __init__(self, indim, hiddim, outdim) -> None:
        super().__init__()
        self.lin1 = torch.nn.Linear(indim, hiddim)
        self.act = torch.nn.ReLU()
        self.lin2 = torch.nn.Linear(hiddim, outdim)

    def forward(self, x, y):
        out = self.lin1(x + y)
        out = self.act(out)
        out = self.lin2(out)
        return out


class Model(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.sm1 = Submod(160, 6400, 6400)
        self.act = torch.nn.GELU()
        self.sm2 = Submod(6400, 12800, 10)

    def forward(self, x):
        out = self.sm1(x, x)
        out = self.act(out)
        out = self.sm2(out, out)
        return out


@pytest.fixture
def setup_model():
    model = DmxModel.from_torch(Model())
    inp = torch.rand((1, 160))
    out_full_0 = model(inp)
    
    # creating transformed submod forward
    DmxModel.create_submod_transform_forward(model, "sm1")
    DmxModel.create_submod_transform_forward(model, "sm2")
    
    return model, inp, out_full_0


def test_unquantized(setup_model):
    model, inp, _ = setup_model
    out1 = model.sm1.transformed_forward(inp, inp)
    out0 = model.sm1.forward(inp, inp)
    # check sm1 output is same between original model and unquantized dmx model
    assert torch.all(
        out1 == out0
    ), "unquantized submodule produced different output from original submodule, should be same"


def test_quantized(setup_model):
    model, inp, out_full_0 = setup_model
    bfp16 = "BFP[8|8]{64}(SN)"
    rules = (
        DmxConfigRule(
            module_types=(dmxnn.Linear,),
            module_config=dict(
                input_format=bfp16,
                weight_format=bfp16,
                bias_format=bfp16,
                output_format=bfp16,
            ),
        ),
    )
    model.configure(None, *rules)
    out1 = model.sm1.transformed_forward(inp, inp)
    out0 = model.sm1.forward(inp, inp)
    # check sm1 output is different between original model and quantized dmx model
    assert torch.any(
        out1 != out0
    ), "quantized submodule produced same output as original submodule, should be different"
    out_full = model(inp)
    out_sub = model.sm1.transformed_forward(inp, inp)
    out_sub = torch.nn.GELU()(out_sub)
    out_sub = model.sm2.transformed_forward(out_sub, out_sub)
    # check model output is same between running whole quantized model vs running quantized submodules step-by-step
    assert torch.all(
        out_full == out_sub
    ), "quantized whole model produced different output from chained quantized submodules, should be same"
    # check model output is different between dmx quantized model and original model
    assert torch.any(
        out_full != out_full_0
    ), "quantized whole model produced same output as unquantized whole model, should be different"



#!/usr/bin/env python3

import torch
from torch.fx import Graph


def qdq_attr(g: Graph, target: str, format: str):
    """Modifies a graph, adds get_attr node with given target,
     and wraps that access in a QdQ pair. It also assumes that
    corresponding scale and zero_point tensors exist.
    """
    attr = g.get_attr(target)
    attr_scale = g.get_attr(target + "_scale")
    attr_zero_point = g.get_attr(target + "_zero_point")
    attr_q = g.call_function(
        torch.ops.dmx.quantize,
        (attr, attr_scale, attr_zero_point, format),
    )
    attr_dq = g.call_function(
        torch.ops.dmx.dequantize, (attr_q, attr_scale, attr_zero_point)
    )
    return attr_dq



from typing import Any, Dict, Tuple
import torch.fx as fx
from torch.fx.node import Argument, Target, Node
from torch.fx.proxy import Proxy

import inspect
from .utils import *


class DMXAwareTransformer(fx.Transformer):
    """
    Substitute as in dmx.model.aware(), replace torch.nn.modules and
    activations with dmx counterpart

    Args:
        module (fx.GraphModule): the module to transform
        node_name_to_scope (dict): A dictionary storing the mapping between node names and scopes

    Attributes:
        module (fx.GraphModule): the module to transform
        node_name_to_scope (dict): A dictionary storing the mapping between node names and scopes
        old_gm (fx.GraphModule): dmxmodules to reuse if module was already transformed
    """

    def __init__(
        self,
        module: fx.GraphModule,
        node_name_to_scope: dict,
        old_gm: fx.GraphModule = None,
        nodeInputs: dict = None,
    ):
        super().__init__(module)
        self.module = module
        self.node_name_to_scope = node_name_to_scope
        self.old_gm = old_gm
        self.dmx_aware_functional_mappings = dmx_aware_functional_mappings.copy()
        self.nodeInputs = nodeInputs

    def add_dmx_aware_functional_mapping(self, target: str, dmx_module_cls):
        self.dmx_aware_functional_mappings[target] = dmx_module_cls

    def call_module(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """
        Execute a ``call_module`` node, replaces the module with its dmx counterpart and returns the result.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the new node and the tracer of the new graph

        """
        assert isinstance(target, str)
        curr_mod = self.module.get_submodule(target)
        node_key = type(curr_mod).__module__ + "." + type(curr_mod).__name__
        if node_key not in dmx_aware_mapping:
            return super().call_module(target, args, kwargs)
        self.add_submod(target, dmx_aware_mapping[node_key].from_raw(curr_mod))
        new_node = self.new_graph.create_node(
            "call_module", target, args=tuple(arg.node for arg in args)
        )
        return Proxy(new_node, self.tracer)

    def add_submod(self, name, mod):
        """
        this function will try to reuse modules in old_gm if possible
        """
        try:
            self.module.add_submodule(name, self.old_gm.get_submodule(name))
        except:
            self.module.add_submodule(name, mod)

    def call_method(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        if target in dmx_aware_method_mapping.keys():
            candidate = target
            curr_name = get_name_for_func_nodes(
                target,
                self.new_graph._graph_namespace._used_names,
                self.new_graph._graph_namespace._base_count,
            )
            scope, _ = self.node_name_to_scope[curr_name]
            new_name = scope + "." + candidate if scope != "" else candidate

            # If new name is not candidate, need to add candidate to used names,
            # otherwise next call_method will use the same candidate. (create_name is also called in create_node)
            if new_name != candidate:
                self.new_graph._graph_namespace.create_name(candidate, None)

            self.add_submod(new_name, dmx_aware_method_mapping[target]())
            new_node = self.new_graph.create_node(
                "call_module",
                new_name,
            )
            new_node.args = process_args(args)
            new_node.kwargs = process_kwargs(kwargs)
            return Proxy(new_node, self.tracer)
        else:
            return super().call_method(target, args, kwargs)

    def create_unique_name_in_scope(self, cand_name):
        # output of get_name_for_func_nodes will replace all . with _
        curr_name = get_name_for_func_nodes(
            cand_name,
            self.new_graph._graph_namespace._used_names,
            self.new_graph._graph_namespace._base_count,
        )
        new_name = (
            cand_name + "_" + curr_name[curr_name.rfind("_") + 1 :]
            if curr_name[-1].isdigit()
            else cand_name
        )
        return new_name

    def call_function(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """
        Execute a ``call_function`` node, replaces the function with its dmx counterpart and returns the result.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the new node and the tracer of the new graph
        """
        assert callable(target)
        node_key = str(target)
        if node_key not in self.dmx_aware_functional_mappings:
            return super().call_function(target, args, kwargs)

        candidate = self.new_graph._target_to_str(target)
        curr_name = get_name_for_func_nodes(
            candidate,
            self.new_graph._graph_namespace._used_names,
            self.new_graph._graph_namespace._base_count,
        )
        curr_target, curr_type = self.node_name_to_scope[curr_name]
        if node_key == "<built-in function add>":
            mod_args, mod_kwargs = self.nodeInputs[curr_name]
            if isinstance(mod_args[0], torch.Tensor) and isinstance(
                mod_args[1], torch.Tensor
            ):
                cand_name = curr_target + ".resadd" if curr_target != "" else "resadd"
            else:
                return super().call_function(target, args, kwargs)
        elif node_key in "<built-in function mul>":
            mod_args, mod_kwargs = self.nodeInputs[curr_name]
            if isinstance(mod_args[0], torch.Tensor) and isinstance(
                mod_args[1], torch.Tensor
            ):
                cand_name = curr_target + ".mul" if curr_target != "" else curr_name
            else:
                return super().call_function(target, args, kwargs)
        elif node_key in [
            repr(eval("torch.matmul")),
            repr(eval("torch.bmm")),
            "<built-in function matmul>",
        ]:
            cand_name = curr_target + ".actmatmul" if curr_target != "" else "actmatmul"
        else:
            cand_name = (
                curr_target + "." + candidate if curr_target != "" else curr_name
            )
        # when cand_name is not the same as curr_name, it did not go through the create_unique_name_in_scope process, so we need to do it here.
        # We also need to add curr_name to used names, otherwise next call_function will use the same curr_name, which will map to the wrong node in the old graph. (create_name is also called in create_node
        if cand_name != curr_name:
            new_name = self.create_unique_name_in_scope(cand_name)
            self.new_graph._graph_namespace.create_name(curr_name, None)
        else:
            new_name = curr_name

        # find out what kwargs to pass in to new module init, which kwargs to pass into forward function of module
        empty_mod = self.dmx_aware_functional_mappings[node_key]()
        accepted_kwarg_keys = inspect.signature(empty_mod.__init__).parameters.keys()
        initkwargs = {}
        newkwargs = {}
        for key, value in kwargs.items():
            if key in accepted_kwarg_keys:
                initkwargs[key] = value
            else:
                newkwargs[key] = value
        newmod = self.dmx_aware_functional_mappings[node_key](**initkwargs)
        newargs, newkwargs = process_args(args), process_kwargs(newkwargs)
        if self.dmx_aware_functional_mappings[
            node_key
        ].is_compound:  # if the module is a compound op
            mod_args, mod_kwargs = self.nodeInputs[curr_name]
            # remove kwargs used for init
            mod_kwargs = {k: v for k, v in mod_kwargs.items() if k not in initkwargs}
            module_graph = newmod.module_graph(*mod_args, **mod_kwargs)
            # last node inserted from merging module_graph
            new_node = self.merge_graph(module_graph, new_name, newargs, newkwargs)

        else:
            self.add_submod(new_name, newmod)
            new_node = self.new_graph.create_node(
                "call_module",
                new_name,
            )
            new_node.args = newargs
            new_node.kwargs = newkwargs
        return Proxy(new_node, self.tracer)

    def merge_graph(self, subgraph, scope, args, kwargs):
        """
        This function is to merge subgraph to graph
        """
        # add submodules of subgraph to self
        for n, m in subgraph.named_children():
            self.add_submod(scope + "." + n, m)

        # add nodes of subgraph to self.new_graph
        node_mapping = {}
        arg_counter = 0
        for node in subgraph.graph.nodes:
            newargs = tuple(
                node_mapping[arg.name] if isinstance(arg, Node) else arg
                for arg in node.args
            )
            newkwargs = {
                k: node_mapping[str(v)] if isinstance(v, Node) else v
                for k, v in node.kwargs.items()
            }
            # placeholder nodes should be skipped and they should be mapped to args/kwargs input to the subgraph
            if node.op == "placeholder":
                if arg_counter < len(args):
                    node_mapping[node.name] = args[arg_counter]
                    arg_counter += 1
                elif node.name in kwargs.keys():
                    node_mapping[node.name] = kwargs[node.name]
                else:
                    raise ValueError("Input to the compound function is incorrect!")

            elif node.op == "call_function" or node.op == "call_method":
                new_node = self.new_graph.create_node(
                    node.op,
                    node.target,
                    newargs,
                    newkwargs,
                    name=scope + "." + node.name,
                )
                node_mapping[node.name] = new_node

            elif node.op == "call_module" or node.op == "get_attr":
                new_node = self.new_graph.create_node(
                    node.op,
                    scope + "." + node.target,
                    newargs,
                    newkwargs,
                )
                node_mapping[node.name] = new_node
        return new_node



import torch
from typing import Optional, Tuple, Dict
from torch.ao.quantization.observer import ObserverBase
from torch.ao.quantization.utils import (
    check_min_max_valid,
    is_per_tensor,
    is_per_channel,
)
import numpy as np
from .format import Format, FixedPoint


def _get_qmin_qmax(dtype: Format) -> Tuple[int, int]:
    if isinstance(dtype, FixedPoint) and dtype.fraction == 0 and dtype.clamp:
        quant_min = -(2 ** (dtype.precision - 1))
        quant_max = 2 ** (dtype.precision - 1) - 1
        if dtype.symmetric:
            quant_min += 1
    else:
        quant_min, quant_max = None, None
    return quant_min, quant_max


class DMXObserverBase(ObserverBase):
    r"""
    Taken from torch.ao.quantization.observer.UniformQuantizationObserverBase
    """

    eps: torch.Tensor

    def __init__(
        self,
        dtype: Format,
        qscheme: torch.qscheme = torch.per_tensor_affine,
        factory_kwargs: Optional[dict] = None,
        eps: float = torch.finfo(torch.float32).eps,
        **kwargs,
    ) -> None:
        assert isinstance(dtype, Format), f"illegal format {dtype}"
        super().__init__(dtype=dtype, **kwargs)
        factory_kwargs = torch.nn.factory_kwargs(factory_kwargs)
        self.qscheme = qscheme
        assert self.qscheme in (
            torch.per_tensor_affine,
            torch.per_tensor_symmetric,
            torch.per_channel_affine,
            torch.per_channel_symmetric,
            torch.per_channel_affine_float_qparams,
        ), "Default Observer only works for per_tensor_affine, \
                per_tensor_symmetric, per_channel_affine, \
                per_channel_symmetric and per_channel_float_qparams quantization scheme"
        self.register_buffer("eps", torch.tensor([eps], **factory_kwargs))
        assert isinstance(
            self.dtype, Format
        ), "Default DMXObserver only works for dmx.compressor.numerical.Format objects as dtype"
        self.quant_min, self.quant_max = _get_qmin_qmax(self.dtype)

    @torch.jit.export
    def _calculate_qparams(
        self, min_val: torch.Tensor, max_val: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""Calculates the quantization parameters, given min and max
        value tensors. Works for both per tensor and per channel cases
        """
        if not check_min_max_valid(min_val, max_val):
            return torch.tensor([1.0], device=min_val.device.type), torch.tensor(
                [0], device=min_val.device.type
            )

        quant_min, quant_max = self.quant_min, self.quant_max
        min_val_neg = torch.min(min_val, torch.zeros_like(min_val))
        max_val_pos = torch.max(max_val, torch.zeros_like(max_val))

        device = min_val_neg.device
        scale = torch.ones(min_val_neg.size(), dtype=torch.float32, device=device)
        zero_point = torch.zeros(min_val_neg.size(), dtype=torch.int64, device=device)

        if (
            self.qscheme == torch.per_tensor_symmetric
            or self.qscheme == torch.per_channel_symmetric
        ):
            max_val_pos = torch.max(-min_val_neg, max_val_pos)
            scale = max_val_pos / (float(quant_max - quant_min) / 2)
            scale = torch.max(scale, self.eps.to(device))
        elif self.qscheme == torch.per_channel_affine_float_qparams:
            scale = (max_val - min_val) / float(quant_max - quant_min)
            self.eps = self.eps.to(device)
            scale = torch.where(scale > self.eps, scale, torch.ones_like(scale))
            # We use the quantize function
            # xq = Round(Xf * inv_scale + zero_point),
            # setting zero_point to (-1 * min *inv_scale) we get
            # Xq = Round((Xf - min) * inv_scale)
            zero_point = -1 * min_val / scale
        else:
            scale = (max_val_pos - min_val_neg) / float(quant_max - quant_min)
            scale = torch.max(scale, self.eps.to(device))
            zero_point = quant_min - torch.round(min_val_neg / scale).to(torch.int)
            zero_point = torch.clamp(zero_point, quant_min, quant_max)

        # For scalar values, cast them to Tensors of size 1 to keep the shape
        # consistent with default values in FakeQuantize.
        if len(scale.shape) == 0:
            # TODO: switch to scale.item() after adding JIT support
            scale = torch.tensor([float(scale)], dtype=scale.dtype, device=device)
        if len(zero_point.shape) == 0:
            # TODO: switch to zero_point.item() after adding JIT support
            zero_point = torch.tensor(
                [int(zero_point)], dtype=zero_point.dtype, device=device
            )
            if self.qscheme == torch.per_channel_affine_float_qparams:
                zero_point = torch.tensor(
                    [float(zero_point)], dtype=zero_point.dtype, device=device
                )

        return scale, zero_point

    def extra_repr(self):
        return f"quant_min = {self.quant_min}, quant_max = {self.quant_max}"


class DummyObserver(DMXObserverBase):
    r"""
    This is a dummy observer that does not do anything
    """

    def __init__(self, dtype: Format, ch_axis: int = -1, **kwargs) -> None:
        super().__init__(dtype=dtype, **kwargs)
        self.dtype = dtype
        self.ch_axis = ch_axis

    def forward(self, x):
        return x

    @torch.jit.export
    def calculate_qparams(self):
        return self._calculate_qparams(torch.empty(0), torch.empty(0))


class MinMaxObserver(DMXObserverBase):
    r"""
    Adapted from torch.ao.quantization.observer.MinMaxObserver,
    supports per-channel
    """

    min_val: torch.Tensor
    max_val: torch.Tensor

    def __init__(
        self,
        dtype: Format = Format.from_shorthand("XP[8,0](CSN)"),
        qscheme: Optional[torch.qscheme] = torch.per_tensor_affine,
        ch_axis: int = -1,
        factory_kwargs: Optional[Dict] = None,
        eps: Optional[float] = torch.finfo(torch.float32).eps,
        **kwargs,
    ) -> None:
        if qscheme == torch.per_channel_affine_float_qparams:
            raise NotImplementedError(
                "MinMaxObserver does not support qscheme: torch.per_channel_affine_float_qparams"
            )
        super().__init__(
            dtype=dtype,
            qscheme=qscheme,
            factory_kwargs=factory_kwargs,
            eps=eps,
            **kwargs,
        )
        self.ch_axis = ch_axis
        factory_kwargs = torch.nn.factory_kwargs(factory_kwargs)
        self.register_buffer("min_val", torch.tensor(float("inf"), **factory_kwargs))
        self.register_buffer("max_val", torch.tensor(float("-inf"), **factory_kwargs))

    def forward(self, x_orig):
        if x_orig.numel() == 0:
            return x_orig
        x = x_orig.detach()  # avoid keeping autograd tape
        dims = list(range(x.dim()))
        x = x.to(self.min_val.dtype)
        if is_per_channel(self.qscheme):
            dims.pop(self.ch_axis)
        min_val_cur = torch.amin(x, dim=dims)
        max_val_cur = torch.amax(x, dim=dims)
        self.min_val = self.min_val.to(min_val_cur.device)
        self.max_val = self.max_val.to(max_val_cur.device)
        min_val = torch.min(min_val_cur, self.min_val)
        max_val = torch.max(max_val_cur, self.max_val)
        if self.min_val.shape:
            self.min_val.copy_(min_val)
            self.max_val.copy_(max_val)
        else:
            self.min_val.data = min_val
            self.max_val.data = max_val
        return x_orig

    @torch.jit.export
    def calculate_qparams(self):
        r"""Calculates the quantization parameters."""
        return self._calculate_qparams(self.min_val, self.max_val)

    @torch.jit.export
    def extra_repr(self):
        return super().extra_repr() + ", min_val = {}, max_val = {}".format(
            self.min_val, self.max_val
        )

    @torch.jit.export
    def reset_min_max_vals(self):
        """Resets the min/max values."""
        self.min_val.copy_(torch.tensor(float("inf")))
        self.max_val.copy_(torch.tensor(float("-inf")))


class HistogramObserver(DMXObserverBase):
    r"""
    Adapted from torch.ao.quantization.observer.HistogramObserver,
    still does not support per-channel
    """

    histogram: torch.Tensor
    min_val: torch.Tensor
    max_val: torch.Tensor

    def __init__(
        self,
        bins: int = 2048,
        upsample_rate: int = 128,
        dtype: Format = Format.from_shorthand("XP[8,0](CSN)"),
        qscheme: Optional[torch.qscheme] = torch.per_tensor_affine,
        ch_axis: int = -1,  # TODO: implement for per-channel
        factory_kwargs: Optional[dict] = None,
        eps=torch.finfo(torch.float32).eps,
        **kwargs,
    ) -> None:
        if not is_per_tensor(qscheme):
            raise NotImplementedError(
                "HistogramObserver's qscheme only support torch.per_tensor_symmetric or torch.per_tensor_affine."
            )
        super().__init__(
            dtype=dtype,
            qscheme=qscheme,
            factory_kwargs=factory_kwargs,
            eps=eps,
            **kwargs,
        )
        factory_kwargs = torch.nn.factory_kwargs(factory_kwargs)
        self.bins = bins
        self.register_buffer("histogram", torch.zeros(self.bins, **factory_kwargs))
        self.register_buffer("min_val", torch.tensor(float("inf"), **factory_kwargs))
        self.register_buffer("max_val", torch.tensor(float("-inf"), **factory_kwargs))
        self.upsample_rate = upsample_rate

    def _get_norm(
        self,
        delta_begin: torch.Tensor,
        delta_end: torch.Tensor,
        density: torch.Tensor,
    ) -> torch.Tensor:
        r"""
        Compute the norm of the values uniformaly distributed between
        delta_begin and delta_end.
        Currently only L2 norm is supported.

        norm = density * (integral_{begin, end} x^2)
             = density * (end^3 - begin^3) / 3
        """
        norm = (
            delta_end * delta_end * delta_end - delta_begin * delta_begin * delta_begin
        ) / 3
        return density * norm

    def _compute_quantization_error(self, next_start_bin: int, next_end_bin: int):
        r"""
        Compute the quantization error if we use start_bin to end_bin as the
        min and max to do the quantization.
        """
        dst_nbins = 2**self.dtype.precision
        bin_width = (self.max_val.item() - self.min_val.item()) / self.bins

        dst_bin_width = bin_width * (next_end_bin - next_start_bin + 1) / dst_nbins
        if dst_bin_width == 0.0:
            return 0.0

        src_bin = torch.arange(self.bins, device=self.histogram.device)
        # distances from the beginning of first dst_bin to the beginning and
        # end of src_bin
        src_bin_begin = (src_bin - next_start_bin) * bin_width
        src_bin_end = src_bin_begin + bin_width

        # which dst_bins the beginning and end of src_bin belong to?
        dst_bin_of_begin = torch.clamp(
            torch.div(src_bin_begin, dst_bin_width, rounding_mode="floor"),
            0,
            dst_nbins - 1,
        )
        dst_bin_of_begin_center = (dst_bin_of_begin + 0.5) * dst_bin_width

        dst_bin_of_end = torch.clamp(
            torch.div(src_bin_end, dst_bin_width, rounding_mode="floor"),
            0,
            dst_nbins - 1,
        )
        dst_bin_of_end_center = (dst_bin_of_end + 0.5) * dst_bin_width

        density = self.histogram / bin_width

        norm = torch.zeros(self.bins, device=self.histogram.device)

        delta_begin = src_bin_begin - dst_bin_of_begin_center
        delta_end = dst_bin_width / 2
        norm += self._get_norm(
            delta_begin,
            torch.ones(self.bins, device=self.histogram.device) * delta_end,
            density,
        )

        norm += (dst_bin_of_end - dst_bin_of_begin - 1) * self._get_norm(
            torch.tensor(-dst_bin_width / 2), torch.tensor(dst_bin_width / 2), density
        )

        dst_bin_of_end_center = dst_bin_of_end * dst_bin_width + dst_bin_width / 2

        delta_begin = -dst_bin_width / 2
        delta_end = src_bin_end - dst_bin_of_end_center
        norm += self._get_norm(torch.tensor(delta_begin), delta_end, density)

        return norm.sum().item()

    def _non_linear_param_search(self) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""Non-linear parameter search.

        An approximation for L2 error minimization for selecting min/max.
        By selecting new min/max, we filter out outliers in input distribution.
        This follows the implementation of NormMinimization::NonlinearQuantizationParamsSearch in
        caffe2/quantization/server/norm_minimization.cc
        """
        assert self.histogram.size()[0] == self.bins, "bins mistmatch"
        bin_width = (self.max_val - self.min_val) / self.bins

        # cumulative sum
        total = torch.sum(self.histogram).item()
        cSum = torch.cumsum(self.histogram, dim=0)

        stepsize = 1e-5  # granularity
        alpha = 0.0  # lower bound
        beta = 1.0  # upper bound
        start_bin = 0
        end_bin = self.bins - 1
        norm_min = float("inf")

        while alpha < beta:
            # Find the next step
            next_alpha = alpha + stepsize
            next_beta = beta - stepsize

            # find the left and right bins between the quantile bounds
            l = start_bin
            r = end_bin
            while l < end_bin and cSum[l] < next_alpha * total:
                l = l + 1
            while r > start_bin and cSum[r] > next_beta * total:
                r = r - 1

            # decide the next move
            next_start_bin = start_bin
            next_end_bin = end_bin
            if (l - start_bin) > (end_bin - r):
                # move the start bin
                next_start_bin = l
                alpha = next_alpha
            else:
                # move the end bin
                next_end_bin = r
                beta = next_beta

            if next_start_bin == start_bin and next_end_bin == end_bin:
                continue

            # calculate the quantization error using next_start_bin and next_end_bin
            norm = self._compute_quantization_error(next_start_bin, next_end_bin)

            if norm > norm_min:
                break
            norm_min = norm
            start_bin = next_start_bin
            end_bin = next_end_bin

        new_min = self.min_val + bin_width * start_bin
        new_max = self.min_val + bin_width * (end_bin + 1)
        return new_min, new_max

    def _adjust_min_max(
        self, combined_min: torch.Tensor, combined_max: torch.Tensor, upsample_rate: int
    ) -> Tuple[torch.Tensor, torch.Tensor, int, int]:
        # We ensure that:
        # (combined_max - combined_min)/(downsample_rate*Nbins) = (max - min)/(upsample_rate*Nbins)
        # This allows us to have a common grid of resolution s, where we can align
        # the input histogram
        # start_idx maps min_val to the histogram bin index.

        hist_bin_width = (self.max_val - self.min_val) / (self.bins * upsample_rate)
        downsample_rate = int(
            torch.ceil(
                (combined_max - combined_min) / (self.bins * hist_bin_width)
            ).item()
        )
        e = downsample_rate * (self.bins * hist_bin_width) - (
            combined_max - combined_min
        )
        # Relax only the max, not the min, so that for one sided distributions, min stays at zero
        combined_max = combined_max + e
        combined_min = combined_min
        start_idx = int(
            torch.round((self.min_val - combined_min) / hist_bin_width).item()
        )
        return combined_min, combined_max, downsample_rate, start_idx

    def _combine_histograms(
        self,
        orig_hist: torch.Tensor,
        new_hist: torch.Tensor,
        upsample_rate: int,
        downsample_rate: int,
        start_idx: int,
        Nbins: int,
    ) -> torch.Tensor:
        # First up-sample the histogram with new data by a factor of L
        # This creates an approximate probability density thats piecwise constant
        upsampled_histogram = new_hist.repeat_interleave(upsample_rate)
        # Now insert the upsampled histogram into the output
        # histogram, which is initialized with zeros.
        # The offset at which the histogram is introduced is determined
        # by the start index as the output histogram can cover a wider range
        histogram_with_output_range = torch.zeros(
            (Nbins * downsample_rate), device=orig_hist.device
        )
        histogram_with_output_range[start_idx : Nbins * upsample_rate + start_idx] = (
            upsampled_histogram
        )
        # Compute integral histogram, double precision is needed to ensure
        # that there are no overflows
        integral_histogram = torch.cumsum(
            histogram_with_output_range, 0, dtype=torch.double
        )[downsample_rate - 1 :: downsample_rate]
        # Finally perform interpolation
        shifted_integral_histogram = torch.zeros((Nbins), device=orig_hist.device)
        shifted_integral_histogram[1:Nbins] = integral_histogram[0:-1]
        interpolated_histogram = (
            integral_histogram - shifted_integral_histogram
        ) / upsample_rate
        orig_hist = orig_hist + interpolated_histogram.to(torch.float)
        return orig_hist

    def forward(self, x_orig: torch.Tensor) -> torch.Tensor:
        if x_orig.numel() == 0:
            return x_orig
        x = x_orig.detach()
        min_val = self.min_val
        max_val = self.max_val
        same_values = min_val.item() == max_val.item()
        is_uninitialized = min_val == float("inf") and max_val == float("-inf")
        if is_uninitialized or same_values:
            min_val, max_val = torch.aminmax(x)
            self.min_val.resize_(min_val.shape)
            self.min_val.copy_(min_val)
            self.max_val.resize_(max_val.shape)
            self.max_val.copy_(max_val)
            assert (
                min_val.numel() == 1 and max_val.numel() == 1
            ), "histogram min/max values must be scalar."
            torch.histc(
                x, self.bins, min=int(min_val), max=int(max_val), out=self.histogram
            )
        else:
            new_min, new_max = torch.aminmax(x)
            combined_min = torch.min(new_min, min_val)
            combined_max = torch.max(new_max, max_val)
            # combine the existing histogram and new histogram into 1 histogram
            # We do this by first upsampling the histogram to a dense grid
            # and then downsampling the histogram efficiently
            (
                combined_min,
                combined_max,
                downsample_rate,
                start_idx,
            ) = self._adjust_min_max(combined_min, combined_max, self.upsample_rate)
            assert (
                combined_min.numel() == 1 and combined_max.numel() == 1
            ), "histogram min/max values must be scalar."
            combined_histogram = torch.histc(
                x, self.bins, min=int(combined_min), max=int(combined_max)
            )
            if combined_min == min_val and combined_max == max_val:
                combined_histogram += self.histogram
            else:
                combined_histogram = self._combine_histograms(
                    combined_histogram,
                    self.histogram,
                    self.upsample_rate,
                    downsample_rate,
                    start_idx,
                    self.bins,
                )

            self.histogram.detach_().resize_(combined_histogram.shape)
            self.histogram.copy_(combined_histogram)
            self.min_val.detach_().resize_(combined_min.shape)
            self.min_val.copy_(combined_min)
            self.max_val.detach_().resize_(combined_max.shape)
            self.max_val.copy_(combined_max)
        return x_orig

    @torch.jit.export
    def calculate_qparams(self):
        is_uninitialized = self.min_val == float("inf") and self.max_val == float(
            "-inf"
        )
        if is_uninitialized:
            return torch.tensor([1.0], device=self.min_val.device.type), torch.tensor(
                [0], device=self.min_val.device.type
            )
        assert self.bins == len(self.histogram), (
            "The number of bins in histogram should be equal to the number of bins "
            "supplied while making this observer"
        )

        new_min, new_max = self._non_linear_param_search()

        return self._calculate_qparams(new_min, new_max)

    def _save_to_state_dict(self, destination, prefix, keep_vars):
        super(HistogramObserver, self)._save_to_state_dict(
            destination, prefix, keep_vars
        )
        destination[prefix + "min_val"] = self.min_val
        destination[prefix + "max_val"] = self.max_val

    def _load_from_state_dict(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs,
    ):
        version = local_metadata.get("version", None)

        if version is None or version < 3:
            # if min_val and max_val are not initialized, update their shape
            # to account for the differences between v2 and v3
            min_val_name, max_val_name = prefix + "min_val", prefix + "max_val"
            if min_val_name in state_dict:
                if state_dict[min_val_name].shape == torch.Size([0]):
                    state_dict[min_val_name] = torch.tensor(float("inf"))
            if max_val_name in state_dict:
                if state_dict[max_val_name].shape == torch.Size([0]):
                    state_dict[max_val_name] = torch.tensor(float("-inf"))

        local_state = ["min_val", "max_val"]
        for name in local_state:
            key = prefix + name
            if key in state_dict:
                val = state_dict[key]
                setattr(self, name, val)
            elif strict:
                missing_keys.append(key)
        super(HistogramObserver, self)._load_from_state_dict(
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys,
            unexpected_keys,
            error_msgs,
        )

    @torch.jit.export
    def extra_repr(self):
        return super().extra_repr() + ", min_val = {}, max_val = {}".format(
            self.min_val, self.max_val
        )


class PercentileObserver(HistogramObserver):
    r"""
    Extending HistogramObserver to allow calculation of percentile from histogram,
    taken from https://github.com/NVIDIA/TensorRT/blob/master/tools/pytorch-quantization/pytorch_quantization/calib/histogram.py#L287
    """

    def __init__(
        self,
        percentile: float = 99.99,
        **kwargs,
    ) -> None:
        if percentile < 0 or percentile > 100:
            raise ValueError(
                "Invalid percentile. Must be in range 0 <= percentile <= 100."
            )
        self.percentile = percentile
        super().__init__(**kwargs)

    def _compute_amax_percentile(self):
        calib_bin_edges = []  # TODO: implement this
        total = self.histogram.sum()
        cdf = np.cumsum(self.histogram / total)
        idx = np.searchsorted(cdf, self.percentile / 100)
        calib_amax = calib_bin_edges[idx]
        calib_amax = torch.tensor(calib_amax.item())  # pylint: disable=not-callable
        return calib_amax

    @torch.jit.export
    def calculate_qparams(self):
        is_uninitialized = self.min_val == float("inf") and self.max_val == float(
            "-inf"
        )
        if is_uninitialized:
            return torch.tensor([1.0], device=self.min_val.device.type), torch.tensor(
                [0], device=self.min_val.device.type
            )
        assert self.bins == len(self.histogram), (
            "The number of bins in histogram should be equal to the number of bins "
            "supplied while making this observer"
        )

        new_min, new_max = self._non_linear_param_search()

        return self._calculate_qparams(new_min, new_max)

    @torch.jit.export
    def extra_repr(self):
        return "percentile = {}".format(
            self.min_val,
        )



import torch
from dmx.compressor import numerical


RANDOM_SEED = 0

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
torch.manual_seed(RANDOM_SEED)


def test_castto_bfp16_1():
    n = 1000
    x = torch.randn((1, n), dtype=torch.float32).to(device)
    x *= 0.5 / x.abs().max()
    x += 1.0

    _x = numerical.CastTo(format="BFP[8|8]{1}(SN)")(x)
    assert torch.allclose(_x, x, rtol=0.0, atol=2**-7)

    x = -x

    _x = numerical.CastTo(format="BFP[8|8]{1}(SN)")(x)
    assert torch.allclose(_x, x, rtol=0.0, atol=2**-7)


def test_bfp16_1_rounding():
    x = torch.Tensor(
        [
            1.0,
            1.0 + 2**-7,
            1.0 + 2**-6,
            1.0 + 2**-6 + 2**-7,
        ]
    ).to(device)
    y = torch.Tensor(
        [
            1.0,
            1.0,
            1.015625,
            1.03125,
        ]
    ).to(device)
    assert torch.all(numerical.CastTo(format="BFP[8|8]{1}(SN)")(x) == y)
    assert torch.all(numerical.CastTo(format="BFP[8|8]{1}(SN)")(-x) == -y)


def test_bfp12_1_rounding():
    x = torch.Tensor(
        [
            1.0,
            1.0 + 2**-3,
            1.0 + 2**-2,
            1.0 + 2**-2 + 2**-3,
        ]
    ).to(device)
    y = torch.Tensor(
        [
            1.0,
            1.0,
            1.25,
            1.5,
        ]
    ).to(device)
    assert torch.all(numerical.CastTo(format="BFP[4|8]{1}(SN)")(x) == y)
    assert torch.all(numerical.CastTo(format="BFP[4|8]{1}(SN)")(-x) == -y)



from typing import Union
import torch
import torch.nn as nn
from .format import Format


class SmoothQuant(nn.Module):
    """
    SmoothQuant is a quantization technique that reduces MatMul quantization error
    by migrating the quantization difficulty from the first input of the MatMul
    (input A) to the second input (input B).

    .. smoothQuant paper:
    https://arxiv.org/pdf/2211.10438.pdf

    Args:
        `a_ch_axis` (int): channel axis for input A of the MatMul
        `b_ch_axis` (int): channel axis for input B of the MatMul
        `a_dynamic` (bool): If set to True, the maximum value of input A will be
            calculated dynamically, default is False.
        `b_dynamic` (bool): If set to True, the maximum value of input B will be
            calculated dynamically, default is False.
        `migration_strength` (float): controls how much quantization difficulty
            we want to migrate from input A to input B, should be between 0 and 1,
            default is 0.5.
        `scale_format` (str or dmx.Format): the numerical format to
            store and compute the scaler, default is "SAME".
        `scale_min`(float): minimum epsilon value used to prevent division by zero
            calculating the scaling factors, default is 1e-5.

    Attributes:
        `a_ch_axis` (int): channel axis for input A of the MatMul
        `b_ch_axis` (int): channel axis for input B of the MatMul
        `a_dynamic` (bool): If set to True, the maximum value of input A will be
            calculated dynamically, default is False.
        `b_dynamic` (bool): If set to True, the maximum value of input B will be
            calculated dynamically, default is False.
        `migration_strength` (float): controls how much quantization difficulty
            we want to migrate from input A to input B, should be between 0 and 1,
            default is 0.5.
        `scale_format` (str or dmx.Format): the numerical format to
            store and compute the scaler, default is "SAME".
        `scale_min` (float): minimum epsilon value used to prevent division by zero
            calculating the scaling factors, default is 1e-5.
        `enabled` (bool): If set to True, smoothQuant will be enabled for both
            input A and input B
        `scale` (Tensor): scaling factors used to scale input A and input B to
            (input A / scale) and (input B * scale), respectively.
        `a_maxabs` (Tensor): the maximum value of absolute of input A
        `b_maxabs` (Tensor): the maximum value of absolute of input B
    """

    calibrating: bool = False

    def __init__(
        self,
        a_ch_axis: int,
        b_ch_axis: int,
        a_dynamic: bool = False,
        b_dynamic: bool = False,
        migration_strength: float = 0.5,
        scale_format: Union[str, Format] = "SAME",
        scale_min: float = 1e-5,
        **kwargs,
    ) -> None:
        super().__init__()
        self.a_ch_axis = a_ch_axis
        self.b_ch_axis = b_ch_axis
        self.register_buffer(
            "a_dynamic", torch.tensor([int(a_dynamic)], dtype=torch.long)
        )
        self.register_buffer(
            "b_dynamic", torch.tensor([int(b_dynamic)], dtype=torch.long)
        )
        self.register_buffer("enabled", torch.tensor([0], dtype=torch.long))
        self.register_buffer("migration_strength", torch.tensor([migration_strength]))
        self.register_buffer("scale_min", torch.tensor([scale_min]))
        self.register_buffer("scale", torch.empty(0))
        self.register_buffer("a_maxabs", torch.empty(0), persistent=False)
        self.register_buffer("b_maxabs", torch.empty(0))

        from .cast import CastTo

        self.scale_cast = CastTo()
        self.set_scale_format(scale_format)

    @torch.jit.export
    def enable(self, enabled: bool = True) -> None:
        """
        Sets/resets the enabled flag.

        Args:
            enabled (bool): if set to True, smoothQuant is enabled, default is True.
        """
        self.enabled[0] = 1 if enabled else 0

    @torch.jit.export
    def disable(self) -> None:
        """
        Disables smoothQuant.
        """
        self.enable(False)

    @torch.jit.export
    def set_dynamic(self, a_dynamic: bool = True, b_dynamic: bool = True) -> None:
        """
        Sets/resets the dynamic flag for inputs A and B.

        Args:
            a_dynamic (bool): if set to True, the maximum value of input A will be
                calculated dynamically, default is True.
            b_dynamic (bool): if set to True, the maximum value of input B will be
                calculated dynamically, default is True.
        """
        self.a_dynamic[0] = 1 if a_dynamic else 0
        self.b_dynamic[0] = 1 if b_dynamic else 0

    @torch.jit.export
    def set_scale_format(self, format: Union[str, Format] = "SAME") -> None:
        """
        Sets/resets the scale_format.

        Args:
            `format` (str or dmx.Format): the numerical format to
        store and compute the scaler, default is "SAME".
        """
        self.scale_cast.set_format(format)

    @torch.jit.export
    def set_migration_strength(self, migration_strength: float) -> None:
        """
        Sets the migration_strength factor.

        Args:
            migration_strength (float): quantization difficulty migration factor,
                should be between 0 and 1, default is 0.5.

        Raises:
            ValueError: If migration_strength is less than 0.0 or greater than 1.0.
        """
        if 0.0 <= migration_strength <= 1.0:
            self.migration_strength[0] = migration_strength
        else:
            raise ValueError(
                f"migration_strength should be between 0 and 1, got {migration_strength}"
            )

    @torch.jit.export
    def reset_scale(self) -> None:
        """
        Resets the scaling tensor to an empty tensor.
        """
        self.scale.data = torch.empty(0)

    @torch.jit.export
    def reset_a_maxabs(self) -> None:
        """
        Resets a_maxabs to an empty tensor.
        """
        self.a_maxabs.data = torch.empty(0)

    @torch.jit.export
    def reset_b_maxabs(self) -> None:
        """
        Resets b_maxabs to an empty tensor.
        """
        self.b_maxabs.data = torch.empty(0)

    @property
    def a_maxabs_exists(self) -> bool:
        """
        Checks if a_maxabs is already calculated.

        Returns:
            True if a_maxabs is calculated, False otherwise.
        """
        return self.a_maxabs.numel() > 0

    @property
    def b_maxabs_exists(self) -> bool:
        """
        Checks if b_maxabs is already calculated.

        Returns:
            True if b_maxabs is calculated, False otherwise.
        """
        return self.b_maxabs.numel() > 0

    def _slicing(self, x: torch.Tensor, dims: torch.Size) -> torch.Tensor:
        """
        Slices the input tensor x such that the size of the sliced output tensor
            matches the given dimensions.

        Args:
            x (Tensor): the input tensor
            dims (Size): the target dimensions

        Returns:
            the output tensor

        Raises:
            RuntimeError: If the dimensions of the input tensor do not match the slicing dimensions.
        """
        if len(x.size()) != len(dims):
            raise RuntimeError(
                "Input tensor should have the same number of dimensions as slicing dimensions"
            )
        slice_str = "["
        for dim in dims:
            slice_str += f"0:{dim},"
        slice_str = slice_str[:-1] + "]"
        return eval(f"x{slice_str}")

    def _padding(self, x: torch.Tensor, dims: torch.Size) -> torch.Tensor:
        """
        Pads the input tensor x such that the size of the padded output tensor
            matches the given dimensions.

        Args:
            x (Tensor): the input tensor
            dims (Size): the target dimensions

        Returns:
            the output tensor

        Raises:
            RuntimeError: If the dimensions of the input tensor do not match the padding dimensions.
        """
        _x_dims = x.size()
        if len(_x_dims) != len(dims):
            raise RuntimeError(
                "Input tensor should have the same number of dimensions as padding dimensions"
            )
        pad_nd_str = "("
        for x_dim, dim in zip(reversed(_x_dims), reversed(dims)):
            pad_nd_str += f"0,{dim - x_dim},"
        pad_nd_str = pad_nd_str[:-1] + ")"
        return nn.functional.pad(x, eval(f"{pad_nd_str}"), "constant", 0)

    def _proper_shape(self, x: torch.Tensor, dim: int) -> torch.Size:
        """
        Calculates the proper shape for the scaling factor of a given tensor.

        Args:
            x: input tensor
            dim: the dimension that scaling should be applied on

        Returns:
            shape of the scaling tensor
        """
        sz = [1] * x.dim()
        sz[dim] = self.scale.numel()
        return torch.Size(sz)

    def scale_a(self, a: torch.Tensor) -> torch.Tensor:
        """
        If smoothQuant is enabled, scales input A.

        Args:
            a (Tensor): input tensor that scaling will be applied on

        Returns:
            scaled input tensor
        """
        if self.enabled[0] == 1:
            sz = self._proper_shape(a, self.a_ch_axis)
            a = a.to(self.scale.device) / self.scale.view(sz)
        return a

    def scale_b(self, b: torch.Tensor) -> torch.Tensor:
        """
        If smoothQuant is enabled, scales input B.

        Args:
            b (Tensor): input tensor that scaling will be applied on

        Returns:
            scaled input tensor
        """
        if self.enabled[0] == 1:
            sz = self._proper_shape(b, self.b_ch_axis)
            b = b.to(self.scale.device) * self.scale.view(sz)
        return b

    def _maxabs(self, x: torch.Tensor, dim: int) -> torch.Tensor:
        """
        Removes the given dim from the input tensor dimension and computes the maximum
        value of each slice of the input tensor in the remaining dimension(s).

        Args:
            x (Tensor): the input tensor
            dim (int): dimension that will be excluded from computing max

        Returns:
            the maximum value of each slice of the input tensor in all dimensions, except the given dim.
        """
        dims = list(range(x.dim()))
        dims.pop(dim)
        return torch.amax(x.abs(), dim=dims)

    def compute_scale(self, a_maxabs: torch.Tensor, b_maxabs: torch.Tensor) -> None:
        """
        Computes the scaling tensor.

        Args:
            a_maxabs (Tensor): the maximum value of absolute of input A
            b_maxabs (Tensor): the maximum value of absolute of input B
        """
        b_maxabs = b_maxabs.to(self.scale_min.device).clamp(
            min=self.scale_min
        )  # to prevent division-by-zero
        _device = self.migration_strength.device
        _scale = (
            (
                (a_maxabs.to(_device) ** self.migration_strength)
                / (b_maxabs.to(_device) ** (1.0 - self.migration_strength))
            )
            .to(self.scale_min.device)
            .clamp(min=self.scale_min)
        )
        self.scale = self.scale_cast(_scale)

    def forward(self, a: torch.Tensor, b: torch.Tensor) -> None:
        """
        Computes the smoothQuant scaling tensor and scales inputs A and B

        Args:
            a (tensor): input tensor A
            b (tensor): input tensor B
        """
        with torch.no_grad():
            if not self.a_maxabs_exists or self.a_dynamic[0] == 1:
                self.a_maxabs = self._maxabs(a, self.a_ch_axis)
            else:
                self.a_maxabs = torch.maximum(
                    self._padding(
                        self._maxabs(a, self.a_ch_axis), self.a_maxabs.size()
                    ),
                    self.a_maxabs,
                )
            if not self.b_maxabs_exists or self.b_dynamic[0] == 1:
                self.b_maxabs = self._maxabs(b, self.b_ch_axis)
            else:
                self.b_maxabs = torch.maximum(
                    self._padding(
                        self._maxabs(b, self.b_ch_axis), self.b_maxabs.size()
                    ),
                    self.b_maxabs,
                )
            self.compute_scale(
                self._slicing(self.a_maxabs, self._maxabs(a, self.a_ch_axis).size()),
                self._slicing(self.b_maxabs, self._maxabs(b, self.b_ch_axis).size()),
            )
            return self.scale_a(a), self.scale_b(b)

    def _load_from_state_dict(self, state_dict, prefix, *args, **kwargs):
        """
        Overrides torch.nn.Module._load_from_state_dict() to avoid tensor shape mismatch
        """
        if prefix + "scale" in state_dict.keys():
            self.scale = state_dict[prefix + "scale"]  # asign scale manually
        super()._load_from_state_dict(state_dict, prefix, *args, **kwargs)

    def extra_repr(self) -> str:
        """
        Returns the extra representation of smoothQuant
        """
        return f"migration_strength = {self.migration_strength.item()}, a_ch_axis = {self.a_ch_axis}, b_ch_axis = {self.b_ch_axis}, scale_format = {self.scale_cast.format}, dynamic = ({self.a_dynamic.bool().item()}, {self.b_dynamic.bool().item()})"


class ActivationWeightSmoothQuant(SmoothQuant):
    """
    This is the derived class for Activation x Weight smoothQuant.

    Args:
        `ch_axis` (int): channel axis for the input activation tensor
        `win_ch_axis` (int): channel axis for the weight tensor
        `migration_strength` (float): controls how much quantization difficulty
            we want to migrate from activations to weights, should be between
            0 and 1, default is 0.5.
        `scale_format` (str or dmx.Format): the numerical format to
            store and compute the scaler, default is "SAME".
        `dynamic` (bool): If set to True, the maximum value of activations will
            be calculated dynamically, default is False.
        `scale_min`(float): minimum epsilon value used to prevent division by
            zero calculating the scaling factors, default is 1e-5.

    Attributes:
        `ch_axis` (int): channel axis for the input activation tensor
        `win_ch_axis` (int): channel axis for the weight tensor
        `fused_to_weight` (bool): If set to True, the scaling factors will be
            fused to the weights, cannot be enabled when dynamic is set.
    """

    def __init__(
        self,
        ch_axis: int,
        win_ch_axis: int,
        migration_strength: float = 0.5,
        scale_format: Union[str, Format] = "SAME",
        dynamic: bool = False,
        scale_min: float = 1e-5,
        **kwargs,
    ) -> None:
        super().__init__(
            a_ch_axis=ch_axis,
            b_ch_axis=win_ch_axis,
            migration_strength=migration_strength,
            scale_format=scale_format,
            a_dynamic=dynamic,
            b_dynamic=False,
            scale_min=scale_min,
            **kwargs,
        )
        self.ch_axis = ch_axis
        self.win_ch_axis = win_ch_axis
        self.register_buffer("fused_to_weight", torch.tensor([0], dtype=torch.long))

    @torch.jit.export
    def set_dynamic(self, dynamic: bool = True) -> None:
        """
        Sets/resets the dynamic flag for the input activation

        Args:
            a_dynamic (bool): if set to True, the maximum value of the input activation
                will be calculated dynamically, default is True.

        Raises:
            RuntimeError: If the ``dynamic`` and the ``fused_to_weight`` flags are both enabled.
        """
        if dynamic and self.fused_to_weight[0] == 1:
            raise RuntimeError(
                "SmoothQuant cannot be dynamic as scale has been fused to weight already"
            )
        super().set_dynamic(a_dynamic=dynamic, b_dynamic=False)

    @torch.jit.export
    def reset_weight_maxabs(self) -> None:
        """
        Resets weight maxabs.
        """
        self.reset_b_maxabs()

    @property
    def dynamic(self) -> torch.Tensor:
        """
        Checks if the dynamic flag is set for the input activation.

        Returns:
            A boolean tensor set to True if the dynamic flag is one,
                and set to False otherwise.
        """
        return self.a_dynamic

    @property
    def weight_maxabs_computed(self) -> bool:
        """
        Checks if weight_maxabs is already calculated.

        Returns:
            True if weight_maxabs is calculated, False otherwise.
        """
        return self.b_maxabs_exists

    @property
    def input_maxabs_exists(self) -> bool:
        """
        Checks if input_maxabs is already calculated.

        Returns:
            True if input_maxabs is calculated, False otherwise.
        """
        return self.a_maxabs_exists

    def scale_weight(self, wgt):
        """
        Scales weight.

        Args:
            wgt (Tensor): the weight tensor that scaling will be applied on

        Returns:
            scaled weight tensor
        """
        return self.scale_b(wgt).to(wgt.device)

    def scale_input(self, inp):
        """
        Scales the input activation.

        Args:
            inp (Tensor): the input tensor that scaling will be applied on

        Returns:
            scaled input activation tensor
        """
        return self.scale_a(inp).to(inp.device)

    def fuse_to_weight(self, wgt: torch.Tensor) -> None:
        """
        Fuses the scaling factor to the weight tensor.

        Args:
            wgt (Tensor): the weight tensor
        """
        wgt.data = self.scale_weight(wgt.data)
        self.fused_to_weight[0] = 1

    def compute_scale(self, inp_maxabs: torch.Tensor) -> None:
        """
        Computes the scaling tensor.

        Args:
            inp_maxabs (Tensor): the maximum value of absolute of input activation
        """
        super().compute_scale(inp_maxabs, self.weight_maxabs)

    def forward(self, inp: torch.Tensor, wgt: torch.Tensor) -> None:
        """
        Computes the smoothQuant scaling tensor and scales input activation and weight

        Args:
            inp (tensor): the input activation tensor
            wgt (tensor): the weight tensor
        """
        with torch.no_grad():
            if not self.weight_maxabs_computed:
                self.weight_maxabs = self._maxabs(wgt, self.win_ch_axis)
            if not self.input_maxabs_exists or self.dynamic[0] == 1:
                self.input_maxabs = self._maxabs(inp, self.ch_axis)
            else:
                self.input_maxabs = torch.maximum(
                    self._maxabs(inp, self.ch_axis), self.input_maxabs
                )
            self.compute_scale(self.input_maxabs)

    def extra_repr(self) -> str:
        """
        Returns the extra representation of Activation x Weight smoothQuant
        """
        return f"migration_strength = {self.migration_strength.item()}, ch_axis = {self.ch_axis}, win_ch_axis = {self.win_ch_axis}, scale_format = {self.scale_cast.format}, dynamic = {self.dynamic.bool().item()}"



import pytest
import torch
from dmx.compressor.modeling import nn as dmxnn

RANDOM_SEED = 0

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
torch.manual_seed(RANDOM_SEED)


@pytest.mark.parametrize("bsz", [1, 8])
@pytest.mark.parametrize("height", [64, 128])
@pytest.mark.parametrize("width", [64, 128])
@pytest.mark.parametrize("channels", [3, 64, 128])
@pytest.mark.parametrize("eps", [1e-5, 1e-12])
@pytest.mark.parametrize("affine", [True, False])
def test_batchnorm2d(bsz, height, width, channels, eps, affine):
    shape = (bsz, channels, height, width)
    x = torch.randn(*shape, device=device).requires_grad_()
    bn0 = torch.nn.BatchNorm2d(
        channels, eps=eps, affine=affine, momentum=0.1, track_running_stats=True
    ).to(device)
    bn1 = dmxnn.BatchNorm2d(
        channels, eps=eps, affine=affine, momentum=0.1, track_running_stats=True
    ).to(device)

    if affine:
        gamma = torch.rand(channels, requires_grad=True, device=device)
        beta = torch.randn(channels, requires_grad=True, device=device)

        bn0.weight.data = gamma.data
        bn0.bias.data = beta.data

        bn1.weight.data = gamma.data
        bn1.bias.data = beta.data

    x0 = x.clone().detach().requires_grad_()
    x1 = x.clone().detach().requires_grad_()

    y0 = bn0(x0)
    y1 = bn1(x1)

    y1.backward(torch.ones_like(x, device=device))
    y0.backward(torch.ones_like(x, device=device))

    atol = 1e-12
    assert torch.allclose(y0, y1, atol=atol)
    assert torch.allclose(x0.grad, x1.grad, atol=atol)



#!/usr/bin/env python3

import torch

from torch.export import ExportGraphSignature, ExportedProgram


from torch.fx import replace_pattern, Transformer

from torch.fx.node import Argument, Node, Target, map_arg, map_aggregate
from torch.fx.proxy import Proxy
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import functools
import itertools
import inspect

import dmx.compressor.numerical.custom_lib
from dmx.compressor.fx.transform import substitute_transform, qDq_transform

from torch.fx.experimental.proxy_tensor import make_fx, get_isolated_graphmodule


def get_workload_graphmodule(wl):
    W = wl()
    x = W.create_batch(1)
    W.model(x["input_ids"])
    m = W.model.body._gm
    return m.to("cpu"), x["input_ids"].to("cpu")


def test_distilgpt2():
    """
    Example:
        >>> result = test_distilgpt2(); print(result.graph.print_tabular())
    """

    from mlreferences import distilgpt2 as wl

    m, x = get_workload_graphmodule(wl)
    qdqm = qDq_transform(m)
    prog = torch.export.export(qdqm, (x,))
    import dmx_compiler
    from dmx_compiler import DMXCompilerConfigs

    config = DMXCompilerConfigs["stablehlo-dmir"]()
    config.use_fx_importer = True
    config.use_tracing = True
    config.use_sharding = False
    config.generate_artifacts = True
    config.decomposition_ops = [
        torch.ops.aten.split.Tensor,
        torch.ops.aten.split_with_sizes,
        torch.ops.aten.t,
    ]

    module = dmx_compiler.compile(qdqm, x, "distilgpt2", config)

    return module


if __name__ == "__main__":
    module = test_distilgpt2()
    print(module.operation.get_asm(large_elements_limit=10, enable_debug_info=False))



import pytest
import torch
from dmx.compressor.modeling import DmxModel

class LinearRelu(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.lin = torch.nn.Linear(50,50)
        self.act = torch.nn.ReLU()
    def forward(self,x):
        return self.act(self.lin(x))

class NW(torch.nn.Module):
    def __init__(self,n_layers):
        super().__init__()
        self.mods = torch.nn.ModuleList([
            LinearRelu() for _ in range(n_layers)])
    def forward(self,x):
        for m in self.mods:
            x = m(x)
        return x
                

@pytest.mark.parametrize("dev_name", ('cpu','cuda'))
@pytest.mark.parametrize("n_layers", (1,3,5))
def test_runtime_measuring(dev_name,n_layers):
    if dev_name == 'cuda' and not torch.cuda.is_available():
        #test again on cpu
        dev_name = 'cpu'
    device = torch.device(dev_name)
    model = NW(n_layers).to(device)

    model = DmxModel.from_torch(model)
    model.to_baseline_mode()
    model(torch.rand(12,50))
    mod_names = ['.'.join(x[0].split('.')[1:]) for x in list(model.named_dmx_modules())]

    with model.measure_runtimes(device,mod_names):
        model(torch.rand(12,50))
    recs = model.get_runtime_records()    
    assert all(x in recs for x in mod_names), \
        'failed to measure the runtime of some modules in baseline mode'

    model.to_basic_mode()
    with model.measure_runtimes(device,mod_names):
        model(torch.rand(12,50))

    recs = model.get_runtime_records()    
    assert all(x in recs for x in mod_names), \
        'failed to measure the runtime of some modules in baseline mode'
    



#!/usr/bin/env python3

# test to make sure the old mltools api is packaged and an alias for dmx.compressor

def test_mltools_alias_dmx_compressor():
    import mltools
    import dmx.compressor

    assert mltools.__file__ == dmx.compressor.__file__



import pytest
import torch
from dmx.compressor.modeling import nn as dmxnn


RANDOM_SEED = 0

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
torch.manual_seed(RANDOM_SEED)


@pytest.mark.parametrize(
    "num_embeddings,embedding_dim",
    (
        (8, 16),
        (16384, 64),
    ),
)
@pytest.mark.parametrize(
    "batch_size",
    (
        1,
        4,
    ),
)
def test_conv2d(
    batch_size,
    num_embeddings,
    embedding_dim,
):
    torch_module = torch.nn.Embedding(num_embeddings, embedding_dim, device=device)
    dmx_module = dmxnn.Embedding(num_embeddings, embedding_dim, device=device)
    dmx_module.weight.data = torch_module.weight.data
    t_inp = torch.randint(0, num_embeddings, (batch_size,), device=device)
    c_inp = t_inp.clone().detach()
    t_out = torch_module(t_inp)
    c_out = dmx_module(c_inp)
    assert torch.allclose(t_out.data, c_out.data)
    assert torch.allclose(torch_module.weight.data, dmx_module.weight.data)



import torch

import torch.nn.functional as F
from dmx.compressor.modeling.nn import ScaledDotProductAttention
from dmx.compressor.modeling import DmxModel


torch.manual_seed(0)
query = torch.rand(2, 100, 1024)
key = torch.rand(2, 100, 1024)
value = torch.rand(2, 100, 1024)
attn_mask = torch.ones(100, 100)
scale = 5


def test_no_kwargs():
    mod = ScaledDotProductAttention()
    gm = mod.module_graph(query, key, value)
    out1 = gm(query, key, value)
    out0 = F.scaled_dot_product_attention(query, key, value)
    assert torch.allclose(out1, out0), "Mismatch at no kwargs"


def test_attn_mask():
    mod = ScaledDotProductAttention()
    gm = mod.module_graph(query, key, value, attn_mask=attn_mask)
    out1 = gm(query, key, value, attn_mask=attn_mask)
    out0 = F.scaled_dot_product_attention(query, key, value, attn_mask=attn_mask)
    assert torch.allclose(out1, out0), "Mismatch at attn_mask"


def test_causal():
    mod = ScaledDotProductAttention()
    gm = mod.module_graph(query, key, value, is_causal=True)
    out1 = gm(query, key, value)
    out0 = F.scaled_dot_product_attention(query, key, value, is_causal=True)
    assert torch.allclose(out1, out0), "Mismatch at is_causal"


def test_scale():
    mod = ScaledDotProductAttention()
    gm = mod.module_graph(query, key, value, scale=scale)
    out1 = gm(query, key, value, scale)
    out0 = F.scaled_dot_product_attention(query, key, value, scale=scale)
    assert torch.allclose(out1, out0, atol=1e-4), "Mismatch at scale"


def test_gqa():
    query = torch.rand(1, 8, 100, 1024)
    key = torch.rand(1, 2, 100, 1024)
    value = torch.rand(1, 2, 100, 1024)
    mod = ScaledDotProductAttention()
    gm = mod.module_graph(query, key, value, enable_gqa=True)
    out1 = gm(query, key, value)
    out0 = F.scaled_dot_product_attention(query, key, value, enable_gqa=True)
    assert torch.allclose(out1, out0), "Mismatch at enable_gqa"


class CustomMod(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.key = torch.nn.Linear(512, 512)
        self.value = torch.nn.Linear(512, 512)
        self.query = torch.nn.Linear(512, 512)

    def forward(self, x, s):
        key = self.key(x)
        value = self.value(x)
        query = self.query(x)
        out = F.scaled_dot_product_attention(query, key, value, scale=s)
        return out


class CustomModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.mods = torch.nn.ModuleList([CustomMod() for _ in range(5)])

    def forward(self, x, s=None):
        for mod in self.mods:
            x = mod(x, s)
        return x


def test_larger_model():
    model = CustomModel()
    x = torch.rand(1, 512)
    with torch.no_grad():
        y1 = model(x)
    model = DmxModel.from_torch(model)
    with torch.no_grad():
        y2 = model(x)
    assert torch.allclose(y1, y2), "Mismatch at custom model"



import torch
from torch.utils.cpp_extension import load
import os

current_path = os.path.dirname(os.path.realpath(__file__))
quant_cpu = load(
    name="quant_cpu",
    sources=[
        os.path.join(current_path, "quant_cpu/quant_cpu.cpp"),
        os.path.join(current_path, "quant_cpu/bit_helper.cpp"),
        os.path.join(current_path, "quant_cpu/sim_helper.cpp"),
    ],
)
if torch.cuda.is_available():
    quant_cuda = load(
        name="quant_cuda",
        sources=[
            os.path.join(current_path, "quant_cuda/quant_cuda.cpp"),
            os.path.join(current_path, "quant_cuda/bit_helper.cu"),
            os.path.join(current_path, "quant_cuda/sim_helper.cu"),
            os.path.join(current_path, "quant_cuda/block_kernel.cu"),
            os.path.join(current_path, "quant_cuda/float_kernel.cu"),
            os.path.join(current_path, "quant_cuda/fixed_point_kernel.cu"),
            os.path.join(current_path, "quant_cuda/quant.cu"),
        ],
    )
else:
    quant_cuda = quant_cpu

__all__ = ["fixed_point_quantize", "block_quantize", "float_quantize"]


def assert_wl_fl(wl, fl, stage=""):
    if wl == -1 and fl != -1:
        raise ValueError("fixed point {} wl {}, fl {}".format(stage, wl, fl))


def get_module(x):
    if x.is_cuda:
        quant_module = quant_cuda
    else:
        quant_module = quant_cpu
    return quant_module


# Modified for research
def fixed_point_quantize(x, wl, fl, clamp=True, symmetric=False, rounding="stochastic"):
    """
    Quantize a single precision Floating Point into low-precision Fixed Point

    Args:
        - :param: `x` (torch.Tensor) :  the single precision number to be quantized
        - :param: `wl` (int) : word length of the fixed point number being simulated
        - :param: `fl` (int) : fractional length of the fixed point number being simulated
        - :param: `clamp` (bool, optional) : clamp input numbers into representable range. if false,
                  the quantization will only simulate the effect on precision
        - :param: `symmetric` (bool, optional) : discard the minimum representable number to make the representable
                  range symmetric
        - :param: `rounding` (string) : rounding mode, \"stochastic\" or \"nearest\" (default: \"stochastic\")

    Returns:
        - a quantized low-precision block floating point number (torch.Tensor)
    """
    assert isinstance(x, torch.Tensor)
    assert rounding in ["stochastic", "nearest", "up", "down"]
    assert_wl_fl(wl, fl)
    quant_module = get_module(x)
    if rounding == "nearest":
        out = quant_module.fixed_point_quantize_nearest(
            x.contiguous(), wl, fl, clamp, symmetric
        )
    elif rounding == "stochastic":
        out = quant_module.fixed_point_quantize_stochastic(
            x.contiguous(), wl, fl, clamp, symmetric
        )
    elif rounding == "up":
        out = quant_module.fixed_point_quantize_up(
            x.contiguous(), wl, fl, clamp, symmetric
        )
    elif rounding == "down":
        out = quant_module.fixed_point_quantize_down(
            x.contiguous(), wl, fl, clamp, symmetric
        )
    return out


def block_quantize(x, wl, dim=-1, symmetric=True, rounding="stochastic"):
    """
    Quantize a single precision Floating Point into low-precision Block Floating Point

    Args:
        - :param: `x` (torch.Tensor) :  the single precision number to be quantized
        - :param: `wl` (int) : word length of the block floating point number being simulated
        - :param: `rounding` (string) : rounding mode, \"stochastic\" or \"nearest\"

    Returns:
        - a quantized low-precision block floating point number (torch.Tensor)
    """
    assert isinstance(
        x, torch.Tensor
    ), "x is not a single precision Floating Point Tensor"
    assert rounding in [
        "stochastic",
        "nearest",
        "down",
        "up",
    ], "invalid rounding mode, {}".format(rounding)
    quant_module = get_module(x)
    if rounding == "nearest":
        out = quant_module.block_quantize_nearest(x.contiguous(), wl, dim, symmetric)
    elif rounding == "stochastic":
        out = quant_module.block_quantize_stochastic(x.contiguous(), wl, dim, symmetric)
    elif rounding == "down":
        out = quant_module.block_quantize_down(x.contiguous(), wl, dim, symmetric)
    elif rounding == "up":
        out = quant_module.block_quantize_up(x.contiguous(), wl, dim, symmetric)
    return out


def float_quantize(x, exp, man, bias=None, flush_subnormal=True, rounding="stochastic"):
    """
    Quantize a single precision Floating Point into low-precision Floating Point

    Args:
        - :attr: `x` (torch.Tensor) : the single precision number(torch.Tensor) to be quantized
        - :attr: `exp` (int) : number of bits allocated for exponent
        - :attr: `man` (int) : number of bits allocated for mantissa, not counting the virtual bit
        - :attr: `bias` (Optional[int]) : exponent bias
        - :attr: `flush_subnormal` (bool) : whether to flush subnormals to zero
        - :attr: `rounding` (string) : rounding mode, \"stochastic\" or \"nearest\"

    Returns:
        - a quantized low-precision floating point number (torch.Tensor)
    """
    assert isinstance(
        x, torch.Tensor
    ), "x is not a single precision Floating Point Tensor"
    assert rounding in ["stochastic", "nearest"], "invalid rounding mode, {}".format(
        rounding
    )
    quant_module = get_module(x)
    if bias is None:
        bias = 2 ** (exp - 1) - 1
    if rounding == "nearest":
        out = quant_module.float_quantize_nearest(
            x.contiguous(), man, exp, bias, flush_subnormal
        )
    elif rounding == "stochastic":
        out = quant_module.float_quantize_stochastic(
            x.contiguous(), man, exp, bias, flush_subnormal
        )
    return out
